import os
import pkg_resources

from flask import Flask, jsonify, make_response, render_template, request
from flask_cors import CORS

from globus.manager.api.blueprints import blueprints
from globus.manager.api.errors import APIBaseError


def negotiate_error_response(request, error):
    if 'text/html' == request.accept_mimetypes.best:
        response = make_response(
            render_template(
                'error.jinja2',
                status_code=error.status_code, **error.to_dict()))
    else:
        response = jsonify(error.to_dict())

    response.status_code = error.status_code

    return response


def handle_validation_error(error):
    return negotiate_error_response(request, error)


class ContentTypeAwareApp(Flask):
    """
    Flask application that allows routes to override routes previously
    registered if the new rule has a mimetype option supplied, and it
    matches the input mimetype. Used to support the Google Drive credential
    post form functionality.
    """
    def __init__(self, name):
        self.views_for_mime_type = {}
        super().__init__(name)

    def add_url_rule(self, rule, endpoint=None, view_func=None,
                     provide_automatic_options=None, **options):
        mt = options.get('mimetype')
        if mt not in self.views_for_mime_type:
            self.views_for_mime_type[mt] = {}
        if rule not in self.views_for_mime_type[mt]:
            self.views_for_mime_type[mt][rule] = {}

        for m in options.get("methods", []):
            self.views_for_mime_type[mt][rule][m] = view_func

        if 'mimetype' in options:
            del options['mimetype']

        super().add_url_rule(
            rule, endpoint=endpoint, view_func=self.content_type_aware_view,
            provide_automatic_options=provide_automatic_options, **options)

    def __lookup_view_func(self, mimetype, url_rule, method):
        """
        Look up the mime type and url_rule in the view list.
        """
        views = self.views_for_mime_type.get(mimetype)
        if views:
            rule = views.get(url_rule.rule)
            if rule:
                return rule.get(method)

    def content_type_aware_view(self, *args, **kwargs):
        mt = request.mimetype
        view_func = None
        if mt:
            view_func = self.__lookup_view_func(
                mt, request.url_rule, request.method)
        if not view_func:
            view_func = self.__lookup_view_func(
                None, request.url_rule, request.method)
        if not view_func:
            raise Exception("unroutable request")

        return view_func(*args, **kwargs)


def make_app(**kwargs):
    app = ContentTypeAwareApp(__name__)

    # Turn on cross-origin support for this app
    CORS(app, max_age=600)

    app.config.from_pyfile('config/default.py')
    app.config.from_envvar('MANAGER_CONFIG', silent=True)
    app.register_error_handler(APIBaseError, handle_validation_error)

    log_level = os.environ.get("GCS_MANAGER_LOG_LEVEL")
    if log_level:
        app.logger.setLevel(log_level)

    # Turn of strict trailing slash enforcement in the underlying
    # Werkzeug routing engine.
    # With strict_slashes = False, http requests with trailing
    # slashes will silently match routes defined without the slash
    # e.g., GET /test/ will match @route('/test').
    app.url_map.strict_slashes = False

    for bp in blueprints:
        app.register_blueprint(bp)

    # Load user mapping modules
    app.config['mapping'] = {}
    for entry_point in (
            pkg_resources.iter_entry_points(
                'globus.manager.mapping')):
        user_mapping = entry_point.load()
        app.config['mapping'][entry_point.name] = user_mapping

    # Load connector modules, passing in the app.
    # They may register additional blueprints with the app.
    app.config['connectors'] = {}
    for entry_point in (
            pkg_resources.iter_entry_points('globus.manager.connectors')):
        connector = entry_point.load()(app)
        app.config['connectors'][connector.CONNECTOR_ID] = connector

    return app
