import logging
import time

import globus_sdk

from cachelib import SimpleCache

from globus.manager.api.authentication.utils import get_auth_client
from globus.manager.api.errors import InternalServerError

log = logging.getLogger(__name__)


cache = SimpleCache()


class AuthenticationState(object):
    """
    Given a token, it will make a call into Globus Auth to introspect the
    token and then assign an authentication state based on the results.
    """
    def __init__(self, token):
        self.token = token
        self.identity_set = []
        self._identity_set_data = None
        self.primary_identity = None
        self._auth_token_execution_count = 0
        self._auth_token_cache_key = 'auth_token_response'

        if token is not None:
            log.debug('Authenticating with token ending in "...{}"'
                      .format(token[-5:]))

            token_info = cache.get(token)
            if token_info is None:
                client = get_auth_client()
                token_info = client.oauth2_token_introspect(
                    token, include='identity_set,session_info')

                cache.set(token, token_info.data, timeout=60)

                # TODO any other authorization on the token introspect data?
                # will need to check scopes when multiple scopes exist.

            self._token_info = token_info

            if token_info.get('active'):
                log.debug('Authentication success.')
                self.primary_identity = token_info.get('sub')
                self.identity_set = token_info.get('identity_set', [])
                self.session_info = token_info.get('session_info', {})
            else:
                log.debug('Authentication failed -- invalid token.')
        else:
            log.debug('Received call with no token.')

    def _get_auth_token(self):
        client = get_auth_client()
        scopes = 'urn:globus:auth:scope:auth.globus.org:view_identities'

        return client.oauth2_client_credentials_tokens(
            requested_scopes=scopes).by_resource_server

    @property
    def is_authenticated(self):
        return self.primary_identity is not None

    @property
    def auth_token(self):
        token_response = cache.get(self._auth_token_cache_key)
        if token_response is None:
            token_response = self._get_auth_token()

            token_exp = token_response['auth.globus.org']['expires_at_seconds']
            cache_expiration_seconds = int((token_exp - time.time()) / 2)
            cache.set(
                self._auth_token_cache_key, token_response,
                timeout=cache_expiration_seconds)

        token_exp = token_response['auth.globus.org']['expires_at_seconds']
        if time.time() > token_exp - 120:
            cache.delete(self._auth_token_cache_key)

            if self._auth_token_execution_count >= 10:
                self._auth_token_execution_count = 0
                raise InternalServerError('Unable to get valid auth token')
            else:
                self._auth_token_execution_count += 1
                return self.auth_token
        else:
            self._auth_token_execution_count = 0
            return token_response['auth.globus.org']['access_token']

    @property
    def identity_set_data(self):
        if self._identity_set_data:
            return self._identity_set_data

        try:
            client = globus_sdk.AuthClient(
                authorizer=globus_sdk.AccessTokenAuthorizer(self.auth_token))
        except globus_sdk.GlobusAPIError as err:
            # If token isn't valid, clear the token from the cache
            # and try again.
            if err.http_status == 401:
                cache.delete(self._auth_token_cache_key)

                client = globus_sdk.AuthClient(
                    authorizer=globus_sdk.AccessTokenAuthorizer(
                        self.auth_token))
            else:
                raise

        r = client.get_identities(ids=self.identity_set)
        self._identity_set_data = r['identities']

        return self._identity_set_data
