import time
from flask import current_app
from globus.manager.api.errors import ForbiddenError


class StorageGatewayAuthorizer():
    """
    This class provides methods to process the storage gateway authorization
    policies
    """
    def __init__(self, storage_gateway):
        self.require_high_assurance = storage_gateway.require_high_assurance
        self.authentication_assurance_timeout = (
            storage_gateway.storage_policies['public'].get(
                'authentication_assurance_timeout', 0))
        self.required_idps = storage_gateway.storage_policies['public'].get(
            'identity_providers', [])
        self.required_domains = storage_gateway.storage_policies['public'].get(
            'domains', [])

    def filter_ha_identities(
            self, identity_set_data, authentications):
        """
        If this is a HA storage gateway, return a copy of identity_set_data
        which only contains items which have been authenticated in the current
        session, otherwise, return the entire set
        """
        if self.require_high_assurance:
            assert self.authentication_assurance_timeout > 0
            assert authentications is not None
            current_app.logger.debug("Filtering idset based on HA policy")
            now = time.time()
            timeout = self.authentication_assurance_timeout * 60

            valid_identity_set_data = []
            for i in identity_set_data:
                session_id_info = authentications.get(i['id'])
                if session_id_info:
                    auth_time = session_id_info.get('auth_time')
                    if auth_time and (now - auth_time) < timeout:
                        valid_identity_set_data.append(i)
                        current_app.logger.debug(
                            "Including id {}, authenticated {} seconds ago"
                            .format(i['id'], now - auth_time))
                    else:
                        current_app.logger.debug(
                            "Excluding id {}, authenticated {} seconds ago"
                            .format(i['id'], now - auth_time))
                else:
                    current_app.logger.debug(
                        "Excluding id {}, not in session".format(i['id']))
        else:
            current_app.logger.debug(
                "No HA policy, including all identities")
            valid_identity_set_data = identity_set_data

        current_app.logger.debug(
            "filter_ha_identities: result id_set = {}"
            .format(valid_identity_set_data))
        return valid_identity_set_data

    def filter_required_domains(self, identity_set_data):
        """
        Return a copy of identity_set_data containing only identities which
        match one of the storage gateway policies required domains
        """
        current_app.logger.debug(
            "Filtering idset based on required domains policy")
        if len(self.required_domains) > 0:
            valid_identity_set_data = []
            for i in identity_set_data:
                username, domain = i.get('username', '@').split('@')
                if len(domain) > 0:
                    if domain.lower() in self.required_domains:
                        current_app.logger.debug(
                            "Including identity {} with username {}".format(
                                i['id'],
                                i['username']))
                        valid_identity_set_data.append(i)
                    else:
                        current_app.logger.debug(
                            "Excluding identity {} with username {}".format(
                                i['id'],
                                i['username']))
        else:
            current_app.logger.debug(
                "No required identity domain, including all identities")
            valid_identity_set_data = identity_set_data

        return valid_identity_set_data

    def filter_required_providers(self, identity_set_data):
        """
        Return a copy of identity_set_data containing only identities which
        match one of the storage gateway policies required id providers
        """
        current_app.logger.debug(
            "Filtering idset based on required providers policy")
        if len(self.required_idps) > 0:
            valid_identity_set_data = []
            for i in identity_set_data:
                if i['identity_provider'] in self.required_idps:
                    current_app.logger.debug(
                        "Including identity {} with username {}".format(
                            i['id'],
                            i['username']))
                    valid_identity_set_data.append(i)
                else:
                    current_app.logger.debug(
                        "Excluding identity {} with username {}".format(
                            i['id'],
                            i['username']))
        else:
            current_app.logger.debug(
                "No required identity providers, including all identities")
            valid_identity_set_data = identity_set_data

        return valid_identity_set_data

    def authorize_identity(
            self, identity_id, identity_set_data,
            authentications=None):
        """
        Return an entry from identity_set_data which matches the policies
        for this storage gateway.  Error string should include
        'current session' to trigger a login in the collections app.
        """
        err_str = 'Identity is not available'
        idset = self.filter_ha_identities(identity_set_data, authentications)

        if len(idset) > 0:
            idset = self.filter_required_providers(idset)
            if len(idset) == 0:
                err_str = 'Identity from a required provider is not available'

        if len(idset) > 0:
            idset = self.filter_required_domains(idset)
            if len(idset) == 0:
                err_str = 'Identity from a required domain is not available'

        for i in idset:
            if i['id'] == identity_id:
                return i

        if self.require_high_assurance:
            raise ForbiddenError(
                "{} in the current session".format(err_str))
        else:
            raise ForbiddenError(
                "{} in the current identity_set".format(err_str))

    def authorize_identities(
            self, identity_set_data,
            authentications=None):
        """
        Return a list of entries from identity_set_data which match the
        storage gateway policies.
        """
        # this is currently only called by policies_local_accounts_get()
        # ensure future callers do not need ha validation
        # idset =
        #  self.filter_ha_identities(identity_set_data, authentications)
        idset = identity_set_data
        idset = self.filter_required_providers(idset)
        idset = self.filter_required_domains(idset)

        return idset
