import globus_sdk

from flask import (
    current_app, Blueprint, g, jsonify, url_for)
from sqlalchemy.orm.exc import NoResultFound

from globus.manager.api.errors import (
    BadRequestError, ForbiddenError, InternalServerError,
    NotImplementedError,
    CredentialDeletedError, ServiceUnavailableError)
from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.authorization import StorageGatewayAuthorizer
from globus.manager.api.models import (
    db, StorageGateway, UserCredential, Collection, Configuration,
    PreallocatedCollection)
from globus.manager.api.schemas import (
    collection_schema)
from globus.manager.api.schemas.request import (
    collection_post_schema, shared_endpoint_schema, endpoint_role_schema)
from globus.manager.api.blueprints.decorators import (
    expects_json, server_timing,
    auth_api, db_read, db_write, transfer_api)
from globus.manager.api.utils import https_endpoint
from globus.manager.api.utils.services import get_transfer_client
from globus.manager.api.blueprints.utils import (
    get_transfer_token, get_dependent_token, make_user_credential)

blueprint = Blueprint('collections', __name__, url_prefix='/api/v1')


@blueprint.route('/collections', methods=['GET'])
@blueprint.route('/collections/<uuid:collection_id>', methods=['GET'])
@authenticated
def collections_get(collection_id=None):
    """Not Implemented."""
    if collection_id is not None:
        # get single
        pass
    else:
        # get all
        pass

    raise NotImplementedError()


@blueprint.route(
    '/collections/<uuid:collection_id>', methods=['PUT', 'DELETE'])
def collections_catchall(collection_id):
    """Not implemented."""
    raise NotImplementedError()


@blueprint.route('/collections', methods=['POST'])
@server_timing
@authenticated
@expects_json(schema=collection_post_schema)
def collections_post():
    """
    POST /collections

    Create a collection.
    """
    timing = g.server_timing
    if g.validated_data.get('user_credential_id') is None:
        try:
            timing.start(db_read)
            storage_gateway = StorageGateway.query.filter_by(
                id=g.validated_data.get('storage_gateway_id')).one()
            timing.finish(db_read)
        except NoResultFound:
            raise BadRequestError('storage_gateway_id does not exist')
        authorizer = StorageGatewayAuthorizer(storage_gateway)

        # confirm the globus identity is requesting this credential for itself
        # and that the identity meets any gateway specific policies
        identity_data = authorizer.authorize_identity(
            g.validated_data['identity_id'], g.auth_state.identity_set_data,
            g.auth_state.session_info.get('authentications', {}))

        # Call connector-specific code to create a credential for
        # this storage gateway. Probably only works with
        # unauthenticated S3
        connector_iface = current_app.config['connectors'].get(
            storage_gateway.connector.id)

        storage_gateway_user_id, display_name, storage_credential = (
            connector_iface.storage_credential(
                g.validated_data, identity_data, storage_gateway))

        # get or create the dummy credential
        timing.start(db_read)
        user_credential = UserCredential.query.filter_by(
            storage_gateway_id=storage_gateway.id,
            storage_gateway_user_id=storage_gateway_user_id,
            identity_id=identity_data['id']).one_or_none()
        timing.finish(db_read)
        if not user_credential:
            user_credential = make_user_credential(
                storage_gateway, identity_data['id'],
                storage_gateway_user_id, display_name, storage_credential)
    else:
        # otherwise get the existing user_credential from the db
        try:
            timing.start(db_read)
            user_credential = UserCredential.query.filter_by(
                id=g.validated_data.get('user_credential_id')).one()
            timing.finish(db_read)
            authorizer = StorageGatewayAuthorizer(
                user_credential.storage_gateway)
        except NoResultFound:
            raise BadRequestError('user_credential_id does not exist')

        # confirm credential has not been soft deleted
        if user_credential.deleted:
            raise CredentialDeletedError(
                'This user credential has been deleted.')

        # confirm the globus identity is requesting this credential for itself
        # and that the identity meets any gateway specific policies
        current_app.logger.error(
            "identity_id={}, identity_set_data={}".format(
                user_credential.identity_id,
                g.auth_state.identity_set_data))
        identity_data = authorizer.authorize_identity(
            user_credential.identity_id, g.auth_state.identity_set_data,
            g.auth_state.session_info.get('authentications', {}))

        connector_iface = current_app.config['connectors'].get(
            user_credential.storage_gateway.connector.id)

    # If this is for a storage gateway that uses local accounts,
    # verify that the user_credential's storage_gateway_user_id is valid
    if hasattr(connector_iface, 'check_user_allowed'):
        local_account_matches = False
        local_account_matches = connector_iface.check_user_allowed(
            user_credential.storage_gateway_user_id,
            user_credential.storage_gateway.storage_policies)
        if not local_account_matches:
            raise ForbiddenError(
                'local identity {} does not match '
                'local account policy'.format(
                    user_credential.storage_gateway_id))

    timing.start(db_read)
    pre_collection = PreallocatedCollection.query.order_by(
        PreallocatedCollection.created_at).first()
    timing.finish(db_read)

    if not pre_collection:
        raise ServiceUnavailableError('No collections available. Try again.')
    else:
        db.session.delete(pre_collection)

    policies = {'public': {'type': 'guest'}}
    ep_default_dir = g.validated_data['default_directory']
    if ep_default_dir != '/':
        policies['private'] = {}
        policies['private']['default_directory'] = ep_default_dir
        ep_default_dir = '/{server_default}/'

    collection = Collection(
        id=pre_collection.id, path=g.validated_data['root_path'],
        user_credential=user_credential, domain=pre_collection.domain,
        collection_policies=policies)

    db.session.add(collection)

    # set any connector specific collection policies
    if hasattr(connector_iface, 'collection_policies_set'):
        connector_iface.collection_policies_set(collection, g.validated_data)

    ep_data = g.validated_data.copy()
    new_ep_metadata = {
        'storage_type':
            user_credential.storage_gateway.connector.display_name,
        'host_endpoint': current_app.config['HOST_ENDPOINT_ID'],
        'gcs_private_uuid': pre_collection.id,
        'https_server': 'https://{}'.format(collection.domain),
        'tlsftp_server': 'tlsftp://{}:443'.format(collection.domain),
        'default_directory': ep_default_dir
    }
    if user_credential.storage_gateway.require_high_assurance:
        new_ep_metadata['authentication_assurance_timeout'] = int(
            user_credential.storage_gateway.storage_policies[
                'public']['authentication_assurance_timeout']) * 60
        new_ep_metadata['force_encryption'] = True

    ep_data.update(new_ep_metadata)
    ep_data, errors = shared_endpoint_schema.dump(ep_data)
    if errors:
        current_app.logger.info(
            'ep_data={}, errors={}'.format(ep_data, errors))

    token = get_transfer_token()
    authorizer = globus_sdk.AccessTokenAuthorizer(token)
    transfer_client = get_transfer_client(authorizer=authorizer)

    def create_shared_endpoint(transfer_client, ep_data):
        timing.start(transfer_api)
        ep_result = transfer_client.create_shared_endpoint(ep_data)
        timing.finish(transfer_api)

        return ep_result['id']

    try:
        new_collection_id = create_shared_endpoint(transfer_client, ep_data)

        if new_collection_id != collection.id:
            db.session.rollback()
            raise InternalServerError(
                'Expected shared endpoint id {} but got {}'.format(
                    collection.id, new_collection_id))

    except globus_sdk.exc.TransferAPIError as ex:
        if ex.http_status == 401:
            token = get_transfer_token(force_new_token=True)
            transfer_client = globus_sdk.TransferClient(
                globus_sdk.AccessTokenAuthorizer(token))

            try:
                create_shared_endpoint(transfer_client, ep_data)
            except globus_sdk.exc.TransferAPIError as ex:
                current_app.logger.exception('Error creating shared endpoint.')
                raise InternalServerError(
                    error=ex.code, error_description=ex.message)
        else:
            current_app.logger.exception('Error creating shared endpoint.')
            raise InternalServerError(
                error=ex.code, error_description=ex.message)

    # commit the deletion of the pre-allocated collection once a share
    # has been made with it's id to prevent future attempt to re-use the
    # id if any of the following steps run into unexpected errors
    db.session.commit()

    def delete_shared_endpoint(transfer_client, ep_id):
        timing.start(transfer_api)
        transfer_client.delete_endpoint(ep_id)
        timing.finish(transfer_api)

    def add_endpoint_role(transfer_client, ep_id, role_data):
        timing.start(transfer_api)
        transfer_client.add_endpoint_role(ep_id, role_data)
        timing.finish(transfer_api)

    try:
        for principal in [user_credential.identity_id,
                          collection.id]:
            role_data, errors = endpoint_role_schema.dump({
                'principal': principal
            })

            add_endpoint_role(transfer_client, collection.id, role_data)
    except globus_sdk.exc.TransferAPIError as ex:
        if ex.http_status == 401:
            token = get_transfer_token(force_new_token=True)
            transfer_client = globus_sdk.TransferClient(
                globus_sdk.AccessTokenAuthorizer(token))

            try:
                add_endpoint_role(
                    transfer_client, collection.id, role_data)
            except globus_sdk.exc.TransferAPIError as ex:
                delete_shared_endpoint(transfer_client, collection.id)

                raise InternalServerError(
                    error=ex.code, error_description=ex.message)
        else:
            delete_shared_endpoint(transfer_client, collection.id)

            raise InternalServerError(
                error=ex.code, error_description=ex.message)

    # set shared endpoint's `owner_string` property to the identity username
    # of the identity in the UserCredential. If this fails, swallow the
    # error and proceed to create the Collection.
    try:
        timing.start(auth_api)
        identities = g.auth_state.identity_set_data
        timing.finish(auth_api)

        identity_username = next(i['username'] for i in identities
                                 if i['id'] == user_credential.identity_id)

        user_transfer_token = get_dependent_token(g.auth_state.token)

        user_transfer_client = globus_sdk.TransferClient(
            globus_sdk.AccessTokenAuthorizer(user_transfer_token))

        timing.start(transfer_api)
        user_transfer_client.update_endpoint(
            collection.id, {'owner_string': identity_username})
        timing.finish(transfer_api)
    except globus_sdk.exc.TransferAPIError as ex:
        if ex.http_status == 401:
            user_transfer_token = get_dependent_token(
                g.auth_state.token, force_new_token=True)
            user_transfer_client = globus_sdk.TransferClient(
                globus_sdk.AccessTokenAuthorizer(user_transfer_token))

            try:
                timing.start(transfer_api)
                user_transfer_client.update_endpoint(
                    collection.id, {'owner_string': identity_username})
                timing.finish(transfer_api)
            except globus_sdk.exc.TransferAPIError as ex:
                current_app.logger.error(
                    'Failed updating owner_string to {} '
                    'on shared endpoint {}.'.format(
                        identity_username, collection.id))
                current_app.logger.debug(ex)
        else:
            current_app.logger.error(
                'Failed updating owner_string to {} '
                'on shared endpoint {}.'.format(
                    identity_username, collection.id))
            current_app.logger.debug(ex)
    except globus_sdk.exc.AuthAPIError:
        current_app.logger.exception(
            'Dependent token grant failed, unable to set owner_string')

    sg = user_credential.storage_gateway
    sp = user_credential.storage_gateway.storage_policies
    if sg.require_high_assurance:
        https_endpoint.HTTPSEndpoint(
            pre_collection.id, current_app.config['CLIENT_ID'],
            current_app.config['CLIENT_SECRET'],
            pre_collection.domain,
            required_domain=sp.get('public', {}).get('domains', [None])[0],
            collection_flags="RequireHighAssurance",
            environment=Configuration.query.filter_by(
                name='environment').one().value.lower(),
            authentication_assurance_timeout=(
                sp['public']['authentication_assurance_timeout'])).setup()
    timing.start(db_write)
    db.session.commit()
    timing.finish(db_write)

    response = jsonify(collection_schema.dump(collection).data)
    response.status_code = 201

    response.headers['Location'] = (
        url_for('.collections_get', collection_id=collection.id))
    response.headers['Server-Timing'] = timing.as_string()

    return response
