import functools

from flask import g, request
from werkzeug.exceptions import HTTPException

from globus.manager.api.errors import (
    BadRequestError, ValidationError, UnsupportedMediaError)
from globus.manager.api.utils.server_timing import ServerTiming


def expects_json(fn=None, schema=None):
    """
    Mark a route as requiring a JSON body and
    optionally validating against a defined schema.

    If a schema is supplied and the request's JSON-body is
    successfully validated, the validated data will be returned
    on the `g` object as `validated_data`.
    """
    if fn is None:
        return functools.partial(expects_json, schema=schema)

    @functools.wraps(fn)
    def decorated_function(*args, **kwargs):
        if not request.is_json:
            raise UnsupportedMediaError()

        try:
            json = request.get_json()
        except HTTPException as ex:
            raise BadRequestError(ex.description)

        g.validated_data = None

        if schema:
            g.validated_data, errors = schema.load(json)

            if errors:
                raise ValidationError(errors=[errors])

        return fn(*args, **kwargs)
    return decorated_function


# Server metrics we want to provide to API consumers
db_read = ('db-read', 'Database Reads')
db_write = ('db-write', 'Database Writes')
transfer_api = ('t-api', 'Transfer API')
auth_api = ('auth-api', 'Auth API')
letsencrypt = ('letsencrypt', 'LetsEncrypt')


def server_timing(fn):
    @functools.wraps(fn)
    def decorated_function(*args, **kwargs):
        g.server_timing = ServerTiming()

        return fn(*args, **kwargs)

    return decorated_function
