from flask import current_app, Blueprint, g, jsonify
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import true

from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.authorization import StorageGatewayAuthorizer
from globus.manager.api.blueprints.decorators import server_timing, db_read
from globus.manager.api.errors import BadRequestError, NotFoundError
from globus.manager.api.models import Connector, StorageGateway
from globus.manager.api.schemas import (
    storage_gateway_policies_schema,
    storage_gateways_policies_schema)

blueprint = Blueprint('policies_blueprint', __name__, url_prefix='/api/v1')


@blueprint.route('/policies', methods=['GET'])
@blueprint.route('/policies/<uuid:storage_gateway_id>', methods=['GET'])
@server_timing
@authenticated
def policies_get(storage_gateway_id=None):
    """
    GET /policies
    GET /policies/:id

    Returns storage policies.
    """
    timing = g.server_timing

    if storage_gateway_id is not None:
        sg_id = str(storage_gateway_id)
        try:
            timing.start(db_read)
            storage_gateway = (StorageGateway.query.filter_by(
                deleted=False, allow_guest_collections=True, id=sg_id)
                .join(StorageGateway.connector)
                .filter(Connector.enabled == true())
                .one())
            timing.finish(db_read)
        except NoResultFound:
            raise NotFoundError()

        response = storage_gateway_policies_schema.dump(storage_gateway).data
    else:
        timing.start(db_read)
        storage_gateways = (StorageGateway.query.filter_by(
            deleted=False, allow_guest_collections=True)
            .join(StorageGateway.connector)
            .filter(Connector.enabled == true())
            .all())
        timing.finish(db_read)

        response = storage_gateways_policies_schema.dump(storage_gateways).data

    resp = jsonify(response)
    resp.headers['Server-Timing'] = timing.as_string()

    return resp


@blueprint.route(
    '/policies/<uuid:storage_gateway_id>/local_accounts', methods=['GET'])
@server_timing
@authenticated
def policies_local_accounts_get(storage_gateway_id):
    """
    GET policies/<uuid:storage_gateway_id>/local_accounts

    Returns a list of usernames for accounts local to the server that
    match the username part of an identity username from the caller's
    identity set.
    """
    timing = g.server_timing

    sg_id = str(storage_gateway_id)
    try:
        timing.start(db_read)
        storage_gateway = (StorageGateway.query.filter_by(
            deleted=False, allow_guest_collections=True, id=sg_id)
            .join(StorageGateway.connector)
            .filter(Connector.enabled == true())
            .one())
        timing.finish(db_read)
    except NoResultFound:
        raise NotFoundError()

    connector_iface = current_app.config['connectors'].get(
        storage_gateway.connector.id)

    mapping = storage_gateway.storage_policies['private'].get('mapping')
    if mapping is None:
        if hasattr(connector_iface, 'default_mapping'):
            mapping = connector_iface.default_mapping

    if mapping is None or not hasattr(connector_iface, 'check_user_allowed'):
        raise BadRequestError(
            error_description='{} storage gateways do not support '
            'local accounts'.format(storage_gateway.connector.display_name))

    # Filter identity set based on storage policies
    authorizer = StorageGatewayAuthorizer(storage_gateway)
    identity_set_data = authorizer.authorize_identities(
        g.auth_state.identity_set_data,
        authentications=g.auth_state.session_info.get(
            'authentications', []))

    # Map local accounts
    local_accounts = []
    for identity_data in identity_set_data:
        try:
            username = mapping(identity_data)
            if connector_iface.check_user_allowed(
                    username, storage_gateway.storage_policies):
                local_accounts.append(username)
        except Exception as e:
            current_app.logger.warning(
                "Connector {} could not map {} to a local user: {}".format(
                    storage_gateway.connector_id,
                    identity_data, e))

    resp = jsonify(local_accounts)
    resp.headers['Server-Timing'] = g.server_timing.as_string()
    return resp
