from flask import (
    current_app, Blueprint, g, jsonify, request, url_for,
    make_response)
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import false, true

from globus.manager.api.errors import (
    BadRequestError, ForbiddenError,
    NotFoundError, NotImplementedError)
from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.authorization import StorageGatewayAuthorizer
from globus.manager.api.models import (
    db, Connector, StorageGateway, UserCredential)
from globus.manager.api.schemas import (user_credential_schema)
from globus.manager.api.schemas.request import (
    user_credential_json_post_schema)
from globus.manager.api.blueprints.decorators import (
    expects_json,
    server_timing,
    db_read,
    db_write)
from globus.manager.api.blueprints.utils import make_user_credential


# connector constants
UNAUTHENTICATED_USER_ID = "UNAUTHENTICATED"

blueprint = Blueprint('user_credentials', __name__, url_prefix='/api/v1')


@blueprint.route('/user_credentials', methods=['GET'])
@server_timing
@authenticated
def user_credentials_get():
    """
    GET /user_credentials
    Returns a list of user credential documents. May be
    filtered by an optional `storage_gateway` query string parameter
    whose value is the `id` of the `storage_gateway`.
    """
    storage_filter = request.args.get('storage_gateway', '')

    query = UserCredential.query.filter(
        UserCredential.identity_id.in_(g.auth_state.identity_set),
        UserCredential.storage_gateway_user_id != UNAUTHENTICATED_USER_ID,
        UserCredential.deleted == false())

    if storage_filter:
        query = query.filter_by(storage_gateway_id=storage_filter)

    g.server_timing.start(db_read)
    query_results = query.all()
    g.server_timing.finish(db_read)

    # XXX: Should this filter credentials based on the policies
    # of the gateway they are associated with?

    results = []
    for cred in query_results:
        connector_iface = current_app.config['connectors'].get(
            cred.storage_gateway.connector_id)
        if hasattr(connector_iface, 'user_credential_get'):
            cred = connector_iface.user_credential_get(cred)
        else:
            cred = user_credential_schema.dump(cred).data
        results.append(cred)

    resp = jsonify(results)
    resp.headers['Server-Timing'] = g.server_timing.as_string()

    return resp


@blueprint.route(
    '/user_credentials/<uuid:user_credential_id>', methods=['GET'])
@server_timing
@authenticated
def user_credentials_get_by_id(user_credential_id):
    """
    GET /user_credentials/:id
    Returns the user credential document specified by the user_credential_id
    """
    uc_id = str(user_credential_id)

    timing = g.server_timing

    timing.start(db_read)
    user_credential = UserCredential.query.filter_by(
        id=uc_id, deleted=False).one_or_none()
    timing.finish(db_read)

    if not user_credential:
        raise NotFoundError()

    if user_credential.identity_id not in g.auth_state.identity_set:
        raise ForbiddenError()

    connector_iface = current_app.config['connectors'].get(
        user_credential.storage_gateway.connector_id)
    if hasattr(connector_iface, 'user_credential_get'):
        user_credential = connector_iface.user_credential_get(user_credential)
    else:
        user_credential = user_credential_schema.dump(user_credential).data
    resp = jsonify(user_credential)
    resp.headers['Server-Timing'] = g.server_timing.as_string()

    return resp


@blueprint.route('/user_credentials', methods=['POST'])
@authenticated
@server_timing
@expects_json(schema=user_credential_json_post_schema)
def user_credentials_post():
    """
    POST /user_credentials/
    Validates given data, then creates a user credential
    Flows differ greatly between content storage gateway types
    """
    post_data = g.validated_data

    # confirm post data is for a valid storage gateway
    try:
        storage_gateway = (StorageGateway.query.filter_by(
            deleted=False, allow_guest_collections=True,
            id=post_data['storage_gateway'])
            .join(StorageGateway.connector)
            .filter(Connector.enabled == true())
            .one())
    except NoResultFound:
        current_app.logger.info(
            'Attempt to create a user credential with an unknown or '
            'disabled storage gateway: {}'
            .format(post_data['storage_gateway']))

        raise BadRequestError('storage_gateway does not exist, is disabled, '
                              'or does not allow guest collections')

    # confirm the globus identity is requesting this credential for itself
    # and that the identity meets any gateway specific policies
    authorizer = StorageGatewayAuthorizer(storage_gateway)
    identity_data = authorizer.authorize_identity(
        post_data['identity_id'], g.auth_state.identity_set_data,
        g.auth_state.session_info.get('authentications', {}))

    # Call out to the connector to do any storage-specific work to
    # determine the credential info
    connector_iface = current_app.config['connectors'].get(
        storage_gateway.connector.id)
    try:
        storage_credential_user_id, display_name, storage_credential = (
            connector_iface.storage_credential(
                post_data, identity_data, storage_gateway))
        user_credential = make_user_credential(
            storage_gateway, post_data['identity_id'],
            storage_credential_user_id,
            display_name=display_name,
            storage_credential=storage_credential)

        if hasattr(connector_iface, 'user_credential_get'):
            user_credential = connector_iface.user_credential_get(
                user_credential)
        else:
            user_credential = user_credential_schema.dump(
                user_credential).data
        response = jsonify(user_credential)
        response.status_code = 201
        response.headers['Location'] = url_for(
            '.user_credentials_get_by_id',
            user_credential_id=user_credential['id'])
        return response
    except AttributeError as e:
        current_app.logger.error(e)
        raise NotImplementedError('connector not implemented.')


@blueprint.route(
    '/user_credentials/<uuid:user_credential_id>', methods=['DELETE'])
@server_timing
@authenticated
def user_credentials_delete(user_credential_id):
    uc_id = str(user_credential_id)

    timing = g.server_timing

    timing.start(db_read)
    user_credential = UserCredential.query.filter_by(
        id=uc_id, deleted=False).one_or_none()
    timing.finish(db_read)

    if not user_credential:
        raise NotFoundError()

    # Verify the caller has access to this credential
    authorizer = StorageGatewayAuthorizer(user_credential.storage_gateway)
    authorizer.authorize_identity(
        user_credential.identity_id, g.auth_state.identity_set_data,
        g.auth_state.session_info.get('authentications', {}))

    # Call out to the connector fo any storage-specific credential cleanup
    connector_iface = current_app.config['connectors'].get(
        user_credential.storage_gateway.connector.id)
    if user_credential.storage_credential:
        try:
            connector_iface.storage_credential_delete(
                user_credential.storage_credential)
            current_app.logger.info(
                'Connector revoked storage credential for ({})'.format(
                    user_credential))
        except AttributeError:
            # OK if a connector doesn't implement this interface
            pass
        except Exception as ex:
            current_app.logger.warning(
                'Connector failed to revoke user credential {} ({})'.format(
                    user_credential, ex))

    # Remove this credential from the database
    user_credential.deleted = True
    user_credential.invalid = True
    user_credential.storage_credential = None

    timing.start(db_write)
    db.session.commit()
    timing.finish(db_write)

    resp = make_response(('', 204))
    resp.headers['Server-Timing'] = timing.as_string()

    return resp
