
from urllib.parse import parse_qsl, urlparse, urlunparse, urlencode
import time

from cachelib import SimpleCache
from flask import current_app, g

import globus_sdk

from globus.manager.api.blueprints.decorators import (
    auth_api, db_read, db_write)
from globus.manager.api.errors import InternalServerError, ResourceExistsError
from globus.manager.api.models import db, UserCredential


cache = SimpleCache()


def get_auth_client(client_id=None, secret=None):
    client_id = client_id or current_app.config['CLIENT_ID']
    secret = secret or current_app.config['CLIENT_SECRET']

    return globus_sdk.ConfidentialAppAuthClient(client_id, secret)


def get_transfer_token(
        force_new_token=False, client_id=None, secret=None,
        _execution_count=1):

    client_id = client_id or current_app.config['CLIENT_ID']
    secret = secret or current_app.config['CLIENT_SECRET']
    cache_key = 'transfer_token_response_{}'.format(client_id)

    if force_new_token:
        cache.delete(cache_key)
        token_response = None
    else:
        token_response = cache.get(cache_key)

    if token_response is None:
        auth_client = get_auth_client(client_id=client_id, secret=secret)

        if g.get('server_timing'):
            g.server_timing.start(auth_api)

        transfer_scope = 'urn:globus:auth:scope:transfer.api.globus.org:all'
        token_response = auth_client.oauth2_client_credentials_tokens(
            requested_scopes=transfer_scope).by_resource_server

        if g.get('server_timing'):
            g.server_timing.finish(auth_api)

        token_exp = \
            token_response['transfer.api.globus.org']['expires_at_seconds']
        cache_expiration_seconds = int((token_exp - time.time()) / 2)
        cache.set(
            cache_key, token_response,
            timeout=cache_expiration_seconds)

    token_exp = token_response['transfer.api.globus.org']['expires_at_seconds']
    if time.time() > token_exp - 120:
        cache.delete(cache_key)

        if _execution_count >= 10:
            raise InternalServerError('Unable to get valid transfer token')
        else:
            return get_transfer_token(_execution_count=_execution_count + 1)

    return token_response['transfer.api.globus.org']['access_token']


def get_dependent_token(token, force_new_token=False, _execution_count=1):
    cache_key = 'dependent_token_response_{}'.format(token)

    if force_new_token:
        cache.delete(cache_key)
        token_response = None
    else:
        token_response = cache.get(cache_key)

    if token_response is None:
        auth_client = get_auth_client()

        if g.get('server_timing'):
            g.server_timing.start(auth_api)

        token_response = auth_client.oauth2_get_dependent_tokens(
            token).by_resource_server

        if g.get('server_timing'):
            g.server_timing.finish(auth_api)

        token_exp = \
            token_response['transfer.api.globus.org']['expires_at_seconds']
        cache_expiration_seconds = int((token_exp - time.time()) / 2)
        cache.set(
            cache_key,
            token_response, timeout=cache_expiration_seconds)

    token_exp = token_response['transfer.api.globus.org']['expires_at_seconds']
    if time.time() > token_exp - 120:
        cache.delete(cache_key)

        if _execution_count >= 10:
            raise InternalServerError('Unable to get valid dependent token')
        else:
            return get_dependent_token(
                token, _execution_count=_execution_count + 1)

    return token_response['transfer.api.globus.org']['access_token']


def update_query_string(url, params, replace=False):
    """
    Given a `url` and a dict of query string `params`, update
    the url's query string to add new params and update existing.
    If `replace=True` any existing query string will be dropped and
    replaced by `params`.
    """
    parsed = urlparse(url)

    if not replace:
        query_string = dict(parse_qsl(parsed.query))
        query_string.update(params)
    else:
        query_string = params

    updated_query_string = urlencode(query_string)

    return urlunparse((parsed.scheme,
                       parsed.netloc,
                       parsed.path,
                       parsed.params,
                       updated_query_string,
                       parsed.fragment))


def make_user_credential(
        storage_gateway, identity_id, storage_gateway_user_id,
        display_name=None, storage_credential=None):
    '''
    helper that makes UserCredential with given data, writes it to the db,
    and returns it.
    if no user credential exists, makes a new one
    if a deleted one exists, un-marks it as deleted and re-uses it
    if a non deleted one exists, raises ResourceExistsError
    '''
    timing = g.server_timing
    timing.start(db_read)
    user_credential = UserCredential.query.filter_by(
        storage_gateway_id=storage_gateway.id,
        storage_gateway_user_id=storage_gateway_user_id,
        identity_id=identity_id).one_or_none()
    timing.finish(db_read)

    if user_credential and not user_credential.deleted:
        current_app.logger.error(
            "make_user_credential: user_credential already exists")
        raise ResourceExistsError()

    if not display_name:
        display_name = (storage_gateway_user_id or
                        storage_gateway.connector.display_name)

    if user_credential and user_credential.deleted:
        # reset
        user_credential.deleted = False
        user_credential.invalid = False
        user_credential.display_name = display_name
        user_credential.storage_credential = storage_credential
    else:
        user_credential = UserCredential(
            display_name=display_name,
            storage_gateway_id=storage_gateway.id,
            storage_gateway_user_id=storage_gateway_user_id or "",
            identity_id=identity_id,
            storage_credential=storage_credential)
        db.session.add(user_credential)

    timing.start(db_write)
    db.session.commit()
    timing.finish(db_write)

    return user_credential
