import re

from flask import current_app
from globus.manager.api.errors import (
    ForbiddenError,
    GatewayTimeout,
)


class BlackPearlLookupError(GatewayTimeout):
    error_description = ("Unable to connect to BlackPearl. This storage "
                         "gateway's storage policies may be misconfigured.")


class BlackPearlConnector(object):
    CONNECTOR_ID = "7e3f3f5e-350c-4717-891a-2f451c24b0d4"
    DISPLAY_NAME = "blackpearl"

    def __init__(self, app):
        self.default_mapping = (
            app.config['mapping']['username_without_domain'])

    def storage_credential(
            self, credential_data, valid_identity_data, storage_gateway):
        user_mapping_func = storage_gateway.storage_policies['private'].get(
            'mapping',
            self.default_mapping)

        username = user_mapping_func(valid_identity_data)

        if username != credential_data['storage_gateway_user_id']:
            raise ForbiddenError(
                'local identity {} does not match '
                'local account policy'.format(
                    credential_data.storage_gateway_user_id))

        # Forbidden if the local username is not allowed by BlackPearl
        if not self.check_user_allowed(
                username, storage_gateway.storage_policies):
            raise ForbiddenError(
                'local identity {} does not match '
                'local account policy'.format(
                    username))

        # Return the requested username as the storage credential
        return username, None, username

    def check_user_allowed(self, username, storage_policies):
        """
        Check if a given username maps to a credential in the BlackPearl
        access id file specified in the storage_policies.
        """
        dsi_config = storage_policies['private'].get('dsi_config', {})
        access_id_file = dsi_config.get('bp_access_id_file')

        # error if the storage policies are misconfigured
        if not access_id_file:
            current_app.logger.error(
                "BlackPearl dsi config is missing bp_access_id_file.")
            raise BlackPearlLookupError()

        for line in open(access_id_file, 'r'):
            if re.search('^[ \t]*' + username + '[ \t]+', line):
                return True

        return False
