import os
import time
import binascii
import requests
import jwt

from cachelib import SimpleCache
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from flask import current_app
from globus.manager.api.errors import GatewayTimeout, ForbiddenError

# cache enterprise access tokens by the EnterpriseID they are for
cache = SimpleCache()


class BoxConnectionError(GatewayTimeout):
    error_description = ("Unable to connect to Box. This storage gateway's "
                         "storage policies may be misconfigured.")


class ClockSkewError(BoxConnectionError):
    error_description = ("Unable to connect to Box. Please ensure your system "
                         "time is correct.")


class BoxConnector(object):
    CONNECTOR_ID = "7c100eae-40fe-11e9-95a3-9cb6d0d9fd63"
    DISPLAY_NAME = "box"

    def __init__(self, app):
        self.default_mapping = (
            app.config['mapping']['username_with_domain'])

    def storage_credential(
            self, credential_data, valid_identity_data, storage_gateway):
        user_mapping_func = storage_gateway.storage_policies['private'].get(
            'mapping',
            self.default_mapping)

        username = user_mapping_func(valid_identity_data)

        if username != credential_data['storage_gateway_user_id']:
            raise ForbiddenError(
                'local identity {} does not match '
                'local account policy'.format(
                    credential_data.storage_gateway_user_id))

        # Forbidden if the local username is not allowed by Box
        # policies
        if not self.check_user_allowed(
                username, storage_gateway.storage_policies):
            raise ForbiddenError(
                'local identity {} does not match '
                'local account policy'.format(
                    username))

        # Return the username as the storage credential
        return (username, None, username)

    def check_user_allowed(self, username, storage_policies):
        """
        Returns True if the /users API call returns a valid
        user after being filtered by the given username
        on the Box enterprise described in the given storage_policies
        """
        access_token = self.get_enterprise_access_token(storage_policies)

        try:
            params = {"filter_term": username}
            headers = {"Authorization": "Bearer %s" % access_token}
            response = requests.get("https://api.box.com/2.0/users",
                                    params=params, headers=headers).json()
            if response.get("total_count") != 1:
                return False

            # sanity check, allow case insensitive match
            return response["entries"][0]["login"].lower() == username
        except Exception as e:
            current_app.logger.error(
                "Unable to get Box user info. Exception: {}".format(e))
            raise BoxConnectionError()

    @staticmethod
    def get_enterprise_access_token(storage_policies):
        """
        Either gets a token by the enterpriseID it is cached under,
        or makes a call to get a fresh token and caches it
        """
        global cache
        enterpriseID = storage_policies[
            "private"].get("dsi_config", {}).get("enterpriseID")

        if not enterpriseID:
            current_app.logger.error("Box config is missing enterpriseID")
            raise BoxConnectionError()

        token = cache.get(enterpriseID)
        if token is None:
            token, expires_in = (
                BoxConnector.get_new_enterprise_access_token(
                    storage_policies))
            cache.set(enterpriseID, token, timeout=expires_in - 60)

        return token

    @staticmethod
    def get_new_enterprise_access_token(storage_policies):
        """
        Uses given storage policies to make a jwt grant
        Returns access_token, expires_in from the response
        """
        dsi_config = storage_policies["private"].get("dsi_config", {})
        enterpriseID = dsi_config.get("enterpriseID")
        boxAppSettings = dsi_config.get("boxAppSettings")
        clientID = boxAppSettings.get("clientID")
        clientSecret = boxAppSettings.get("clientSecret")
        appAuth = boxAppSettings.get("appAuth", {})
        publicKeyID = appAuth.get("publicKeyID")
        privateKey = appAuth.get("privateKey")
        passphrase = appAuth.get("passphrase")
        authentication_url = "https://api.box.com/oauth2/token"

        # error if the storage policies are misconfigured
        if not (enterpriseID and clientID and clientSecret and publicKeyID and
                privateKey and passphrase):
            # current_app.logger.error("Box config is missing values")
            raise BoxConnectionError()

        # decrypt private key
        try:
            key = load_pem_private_key(
                data=privateKey.encode("utf8"),
                password=passphrase.encode("utf8"),
                backend=default_backend())
        except Exception as e:
            current_app.logger.error(
                "Unable to decrypt Box privateKey. Exception: {}".format(e))
            raise BoxConnectionError()

        # contruct JWT assertion
        try:
            claims = {"iss": clientID,
                      "sub": enterpriseID,
                      "box_sub_type": "enterprise",
                      "aud": authentication_url,
                      "jti": binascii.hexlify(os.urandom(64)).decode('utf-8'),
                      "exp": int(round(time.time(), 0) + 30)}
            assertion = jwt.encode(claims, key, algorithm="RS512",
                                   headers={"kid": publicKeyID})
        except Exception as e:
            current_app.logger.error(
                "Unable to encode JWT assertion. Exception: {}".format(e))
            raise BoxConnectionError()

        # Make the request, parse the JSON,
        # and extract the access token
        try:
            data = {
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": assertion,
                "client_id": clientID,
                "client_secret": clientSecret
            }
            response = requests.post(authentication_url, data=data).json()
            access_token = response.get("access_token")
            expires_in = response.get("expires_in", 0)

        except Exception as e:
            current_app.logger.error(
                "Unable to make token request to Box. Exception: {}".format(
                    e))
            raise BoxConnectionError()

        if access_token is None:
            current_app.logger.error(
                "Response has no access token: {}".format(response))
            if "the 'exp' claim" in response.get(
                    "error_description", ""):
                raise ClockSkewError()
            else:
                raise BoxConnectionError()

        return access_token, expires_in
