import boto
import json
import socket

from flask import current_app
from boto.s3.connection import OrdinaryCallingFormat
from boto.exception import S3ResponseError
from globus.manager.api.errors import (
    ForbiddenError,
    GatewayTimeout,
)


class CephConnectionError(GatewayTimeout):
    error_description = ("Unable to connect to Ceph. This storage gateway's "
                         "storage policies may be misconfigured.")


class CephConnector(object):
    CONNECTOR_ID = "1b6374b0-f6a4-4cf7-a26f-f262d9c6ca72"
    DISPLAY_NAME = "ceph"

    def __init__(self, app):
        self.default_mapping = (
            app.config['mapping']['username_without_domain'])

    def storage_credential(
            self, credential_data, valid_identity_data, storage_gateway):
        user_mapping_func = storage_gateway.storage_policies['private'].get(
            'mapping',
            self.default_mapping)

        username = user_mapping_func(valid_identity_data)

        if username != credential_data['storage_gateway_user_id']:
            raise ForbiddenError(
                'local identity {} does not match '
                'local account policy'.format(
                    credential_data.storage_gateway_user_id))

        # Forbidden if the local username is not allowed by Ceph
        if not self.check_user_allowed(
                username, storage_gateway.storage_policies):
            raise ForbiddenError(
                'local identity {} does not match '
                'local account policy'.format(
                    username))

        # Return the requested username as the storage credential
        return username, None, username

    def check_user_allowed(self, username, storage_policies):
        """
        Check if a given username maps to a credential set on the Ceph
        gateway described in the given storage_policies
        """
        dsi_config = storage_policies['private'].get('dsi_config', {})
        s3_endpoint = dsi_config.get('s3_endpoint')
        key_id = dsi_config.get('ceph_admin_key_id')
        secret_key = dsi_config.get('ceph_admin_secret_key')

        # error if the storage policies are misconfigured
        if not (s3_endpoint and key_id and secret_key):
            current_app.logger.error(
                "Ceph dsi config is missing s3_endpoint, "
                "ceph_admin_key_id or ceph_admin_secret_key.")
            raise CephConnectionError()

        # extract http:// or https:// defaulting to https if not present
        if s3_endpoint.startswith('http://'):
            use_ssl = False
            host_and_port = s3_endpoint[len('http://'):]
        elif s3_endpoint.startswith('https://'):
            use_ssl = True
            host_and_port = s3_endpoint[len('https://'):]
        else:
            current_app.logger.error("Neither http nor https found in "
                                     "s3_endpoint {}".format(s3_endpoint))
            raise CephConnectionError()

        # get host name and port number if any, defaulting to port 443
        split = host_and_port.split(':')
        host = split[0]
        try:
            port = int(split[1])
        except IndexError:
            if use_ssl:
                port = 443
            else:
                port = 80
            current_app.logger.debug(
                "No port found in Ceph s3_endpoint {}, "
                "defaulting to {}".format(s3_endpoint, port))
        except ValueError:
            current_app.logger.error(
                "Port {} from s3_endpioint {} is not a valid port".format(
                    split[1], s3_endpoint))
            raise CephConnectionError()

        # clear any existing config to prevent conflicts, then set our own
        # values
        for section in boto.config.sections():
            boto.config.remove_section(section)
        boto.config.add_section('Boto')
        boto.config.set('Boto', 'num_retries', '2')  # don't spin too long
        boto.config.add_section('s3')
        boto.config.set('s3', 'use-sigv4', 'True')

        # create the connection state for the admin bucket
        conn = boto.connect_s3(host=host, port=port, is_secure=use_ssl,
                               aws_access_key_id=key_id,
                               aws_secret_access_key=secret_key,
                               calling_format=OrdinaryCallingFormat())

        conn.auth_region_name = 'us-east-1'  # not actually used
        bucket = conn.get_bucket('/admin', validate=False)
        current_app.logger.debug(
            "About to connect to Ceph host {} "
            "on port {} and {} using ssl".format(
                host, port, "not" if not use_ssl else ""))

        # try to get info for the given username
        try:
            res = bucket.get_subresource('uid={}'.format(username), 'user')

            # sanity check that we got the expected user_id back
            if json.loads(res)['user_id'] != username:
                current_app.logger.error(
                    "Unexpected Ceph user id mismatch {} != {}".format(
                        username, res['user_id']))
                return False

            current_app.logger.debug("Ceph user {} was found".format(username))
            return True

        # we expect to get a 404 for a username that does not map
        except S3ResponseError as e:
            if e.status == 404:
                current_app.logger.debug(
                    "Ceph user {} not found".format(username))
                return False
            else:
                current_app.logger.error(
                    "S3 Response Error on Ceph connection: {}".format(e))
                raise CephConnectionError()
        except socket.error as e:
            current_app.logger.error(
                "Socket error on Ceph connection: {}".format(e))
            raise CephConnectionError()
