import httplib2
import jwt

from flask import current_app, g, redirect, request, url_for, Blueprint
from oauth2client import client as google_oauth
from globus.manager.api.models import (
    db,
    Configuration,
    Connector,
    StorageGateway,
    UserCredential,
)
from globus.manager.api.authentication.decorators import authenticated
from globus.manager.api.authorization import StorageGatewayAuthorizer
from globus.manager.api.blueprints.utils import update_query_string
from globus.manager.api.errors import (
    BadRequestError,
    InternalServerError,
    NotImplementedError,
    ValidationError,
)
from globus.manager.api.schemas.request import (
    user_credential_form_post_schema,
)
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import true


class GoogleCloudPlatform(object):
    __blueprint_name = 'google_cloud_platform'
    __connectors = {}

    def __init__(self, app):
        blueprint = Blueprint(
            GoogleCloudPlatform.__blueprint_name,
            __name__,
            url_prefix='/api/v1')
        blueprint.add_url_rule(
            '/authcallback_google',
            'authcallback_google',
            self.authcallback_google,
            methods=['GET'])
        blueprint.add_url_rule(
            '/user_credentials',
            'user_credentials_post',
            self.user_credentials_post,
            methods=['POST'],
            mimetype="application/x-www-form-urlencoded")

        if GoogleCloudPlatform.__blueprint_name not in app.blueprints.keys():
            app.register_blueprint(blueprint)

    def add_connector(self, connector_id, scopes):
        GoogleCloudPlatform.__connectors[connector_id] = scopes

    def storage_credential_delete(self, storage_credential):
        token = storage_credential.get('refresh_token')
        if token:
            gcred = google_oauth.GoogleCredentials(
                access_token=None, client_id=None, client_secret=None,
                refresh_token=token, token_expiry=None, token_uri=None,
                user_agent=None)
            gcred.revoke(httplib2.Http())

    @authenticated
    def user_credentials_post(self):
        post_data, errors = user_credential_form_post_schema.load(request.form)

        if errors:
            raise ValidationError(errors=[errors])
        # confirm post data is for a valid storage gateway
        try:
            storage_gateway = (StorageGateway.query.filter_by(
                deleted=False,
                allow_guest_collections=True,
                id=post_data['storage_gateway'])
                .join(StorageGateway.connector)
                .filter(Connector.enabled == true())
                .one())
        except NoResultFound:
            current_app.logger.info(
                'Attempt to create a user credential with an unknown or '
                'disabled storage gateway: {}'
                .format(post_data['storage_gateway']))
            raise BadRequestError(
                'storage_gateway does not exist, is '
                'disabled, or does not allow guest collections')

        if storage_gateway.connector.id not in self.__connectors:
            raise NotImplementedError()

        authorizer = StorageGatewayAuthorizer(storage_gateway)
        authorizer.authorize_identity(
            post_data['identity_id'],
            g.auth_state.identity_set_data,
            authentications=g.auth_state.session_info.get('authentications'))

        app_secret = Configuration.query.filter_by(name='secret').one().value
        state = jwt.encode(post_data, app_secret, algorithm='HS256')

        # Set up the oauth2 flow.
        # access_type is `offline` by default.
        # response_type is `code` by default.
        # prompt must be set to `consent` to get a refresh token,
        # even when access_type is offline.
        flow = google_oauth.OAuth2WebServerFlow(
            state=state,
            client_id=storage_gateway.credential['client_id'],
            client_secret=storage_gateway.credential['secret'],
            scope=GoogleCloudPlatform.__connectors[
                storage_gateway.connector.id],
            redirect_uri=url_for(
                ".".join([
                    GoogleCloudPlatform.__blueprint_name,
                    'authcallback_google']),
                _external=True, _scheme='https'),
            prompt='consent')

        login_hint = post_data.get('login_hint', '')
        if login_hint:
            flow.params['login_hint'] = login_hint
            flow.params['hd'] = login_hint.split('@')[1]

        # respond with a redirect
        return redirect(flow.step1_get_authorize_url(), code=303)

    def authcallback_google(self):
        state = request.args.get('state')

        if not state:
            # Every request to this route MUST have a state parameter.
            raise BadRequestError('missing state parameter')

        app_secret = Configuration.query.filter_by(name='secret').one().value

        try:
            state_data = jwt.decode(state, app_secret, algorithms=['HS256'])
        except jwt.exceptions.InvalidTokenError:
            raise BadRequestError('Invalid state parameter')

        error = request.args.get('error')
        if error:
            # Google sent us back with an error. Send the user-agent
            # back to the `redirect_uri` that was provided in the original
            # POST to /user_credentials + the error as an added query param.
            redirect_to = update_query_string(
                state_data['redirect_uri'], dict(error=error))

            current_app.logger.error(
                'authcallback_google error: {}'.format(error))

            return redirect(redirect_to)

        try:
            storage_gateway = StorageGateway.query.filter_by(
                deleted=False, allow_guest_collections=True,
                id=state_data['storage_gateway']).one()
        except NoResultFound:
            current_app.logger.error(
                'Made it to /authcallback_google with an unknown '
                'storage_gateway {}.'.format(state_data['storage_gateway']))
            raise BadRequestError('storage_gateway does not exist')

        code = request.args.get('code')
        if code:
            # If we have a `code` we can proceed to exchange it for
            # an access token.
            flow = google_oauth.OAuth2WebServerFlow(
                client_id=storage_gateway.credential['client_id'],
                client_secret=storage_gateway.credential['secret'],
                scope=GoogleCloudPlatform.__connectors[
                    storage_gateway.connector.id],
                redirect_uri=url_for(
                    ".".join([
                        GoogleCloudPlatform.__blueprint_name,
                        'authcallback_google']),
                    _external=True, _scheme='https'))

            try:
                credentials = flow.step2_exchange(code)
            except google_oauth.FlowExchangeError as ex:
                # FlowExchangeError will be the `error` and `error_description`
                # fields concantenated together, so you might see something
                # like `invalid_grantCode was already redeemed`.
                current_app.logger.exception(
                    'Error during Google authentication flow')
                redirect_to = update_query_string(
                    state_data['redirect_uri'],
                    dict(error='flow_exchange_error', error_description=ex))

                return redirect(redirect_to)

            # validate token response
            validated_email = credentials.id_token['email']

            allowed_domains = storage_gateway.storage_policies['public'].get(
                'domains', [])

            if (allowed_domains and
                    validated_email.split('@')[1] not in allowed_domains):
                redirect_to = update_query_string(
                    state_data['redirect_uri'],
                    dict(error='invalid_domain',
                         error_description='domain not allowed'))

                return redirect(redirect_to)

            google_sub = credentials.id_token['sub']
            user_credential = UserCredential.query.filter_by(
                storage_gateway_id=storage_gateway.id,
                storage_gateway_user_id=google_sub,
                identity_id=state_data['identity_id']).one_or_none()

            if user_credential:
                # update existing
                user_credential.display_name = credentials.id_token['email']
                user_credential.storage_credential = {
                    'refresh_token': credentials.refresh_token,
                    'email': credentials.id_token['email'],
                    'scopes': list(credentials.scopes)
                }
                user_credential.deleted = False
                user_credential.invalid = False

                log_message = 'Updated existing {}'
            else:
                # create new user credential
                user_credential = UserCredential()
                user_credential.storage_gateway_user_id = \
                    credentials.id_token['sub']
                user_credential.display_name = credentials.id_token['email']
                user_credential.identity_id = state_data['identity_id']
                user_credential.storage_gateway = storage_gateway
                user_credential.storage_credential = {
                    'refresh_token': credentials.refresh_token,
                    'email': credentials.id_token['email'],
                    'scopes': list(credentials.scopes)

                }

                db.session.add(user_credential)

                log_message = 'Added new {}'
            try:
                db.session.commit()
            except Exception:
                current_app.logger.exception('commit() failed')
                db.session.rollback()

                raise InternalServerError()

            current_app.logger.info(log_message.format(user_credential))
            redirect_to = update_query_string(
                state_data['redirect_uri'],
                dict(user_credential_id=user_credential.id))

            return redirect(redirect_to)

        # if the request has neither an `error` or a `code` it's a bad request.
        raise BadRequestError()
