import httplib2
import json

import oauth2client
from flask import current_app

from globus.manager.api.blueprints.utils import update_query_string
from globus.manager.api.errors import (
    BadRequestError, ForbiddenError, InternalServerError
)
from globus.manager.api.models import db
from globus.manager.api.schemas import ma
from globus.manager.api.schemas.user_credential import UserCredentialSchema

from globus.manager.api.connectors.google_cloud_platform import (
    GoogleCloudPlatform,
)


class GoogleCloudStorageProjectSchema(ma.Schema):
    projectId = ma.Str()
    name = ma.Str()


class GoogleCloudStorageUserCredentialSchema(UserCredentialSchema):
    projects = ma.List(
        ma.Nested(GoogleCloudStorageProjectSchema),
        attribute="storage_credential.projects")


google_cloud_user_credential_schema = (
    GoogleCloudStorageUserCredentialSchema())


class GoogleCloudStorageConnector(object):
    CONNECTOR_ID = "56366b96-ac98-11e9-abac-9cb6d0d9fd63"
    DISPLAY_NAME = "Google Cloud Storage"

    def __init__(self, app):
        self.google_cloud_platform = GoogleCloudPlatform(app)

        SCOPES = (
            'email '
            'https://www.googleapis.com/auth/devstorage.read_write '
            'https://www.googleapis.com/auth/cloudplatformprojects.readonly '
        )
        self.google_cloud_platform.add_connector(
            self.CONNECTOR_ID,
            SCOPES)

    def storage_credential_delete(self, storage_credential):
        self.google_cloud_platform.storage_credential_delete(
            storage_credential)

    def user_credential_get(self, user_credential):
        refresh_token = user_credential.storage_credential['refresh_token']
        secret = user_credential.storage_gateway.credential['secret']
        client_id = user_credential.storage_gateway.credential['client_id']

        cred = oauth2client.client.GoogleCredentials(
            None, client_id, secret,
            refresh_token, None, "https://accounts.google.com/o/oauth2/token",
            "globus-connect-server-manager")
        http = cred.authorize(httplib2.Http())

        final_page = False
        page_token = None
        projects = []
        while not final_page:
            url = (
                "https://cloudresourcemanager.googleapis.com/v1/projects")
            if page_token:
                url = update_query_string(
                    url,
                    dict(
                        page_token=page_token,
                        fields="nextPageToken,projects(projectId,name))"))
            try:
                response, content = http.request(url)
                status = int(response.get('status', 0))
            except oauth2client.client.HttpAccessTokenRefreshError as e:
                status = 400
                response = str(e)
            if status == 200:
                current_app.logger.debug("content={}".format(content))
                data = json.loads(content.decode("utf8"))
                page_token = data.get('nextPageToken')
                final_page = (page_token is None)
                projects.extend(data.get('projects', []))
            # if we get back a 4xx error, the credential is no longer valid
            elif status >= 400 and status < 500:
                current_app.logger.error(
                    "{} error from Google while attempting to list projects, "
                    "marking user_credential as invalid".format(response))
                user_credential.invalid = True
                db.session.commit()
                projects = []
                break
            # other errors could be temporary so we don't invalidate the cred
            else:
                current_app.logger.error(
                    "{} error from Google while attempting to list projects"
                    .format(response))
                projects = []
                break

        # If google_cloud_storage_projects in the DSI config, filter out
        # any projects that are not in that list
        dsi_config = (
            user_credential.storage_gateway.storage_policies.get(
                'private', {}).get('dsi_config', {}))
        allowed_projects = dsi_config.get('google_cloud_storage_projects')
        if allowed_projects:
            projects = list([
                p for p in projects
                if p['projectId'] in allowed_projects
            ])

        sc = user_credential.storage_credential
        sc['projects'] = projects
        user_credential.storage_credential = sc

        # Dump using our own schema
        user_credential, errs = (
            google_cloud_user_credential_schema.dump(
                user_credential))
        if errs:
            current_app.logger.error(
                "Error serializing user_credential: {}".format(errs))
            raise InternalServerError(
                "Error processing Google projects: {}".format(errs))
        current_app.logger.error("credential={}".format(user_credential))

        return user_credential

    def collection_policies_set(self, collection, data):
        """
        Given a collection and validated data for a collection post,
        - Confirms the data has the required google_cloud_storage_project
        - Confirms the project is allowed by the collection's storage gateway
        - Confirms the project is viewable by the collection's user credential
        - Sets the collection's collection_policies to include the project
        """
        user_credential = collection.user_credential
        project = data.get("google_cloud_storage_project")
        if not data.get("google_cloud_storage_project"):
            raise BadRequestError(
                "Collections on gateways using the Google Cloud Storage "
                "connector require the google_cloud_storage_project field.")

        # confirm project is allowed by storage gateway
        dsi_config = user_credential.storage_gateway.storage_policies.get(
            'private', {}).get('dsi_config', {})
        allowed_projects = dsi_config.get('google_cloud_storage_projects')
        if allowed_projects and project not in allowed_projects:
            raise ForbiddenError("This Google Cloud Storage project is not "
                                 "allowed by the Storage Gateway")

        # confirm project is accessible by user_credential
        refresh_token = user_credential.storage_credential['refresh_token']
        secret = user_credential.storage_gateway.credential['secret']
        client_id = user_credential.storage_gateway.credential['client_id']
        cred = oauth2client.client.GoogleCredentials(
            None, client_id, secret,
            refresh_token, None, "https://accounts.google.com/o/oauth2/token",
            "globus-connect-server-manager")
        http = cred.authorize(httplib2.Http())
        url = "https://cloudresourcemanager.googleapis.com/v1/projects/{}"
        response, content = http.request(url.format(project))

        if int(response.get('status', 0)) != 200:
            raise ForbiddenError(
                "The given Google project is not accessible by "
                "the given user_credential")
        else:
            # sanity check we got back the expected project
            data = json.loads(content.decode("utf8"))
            if data.get("projectId") != project:
                raise InternalServerError("Unexpected Google Cloud response")

        # set collection's collection_policies
        existing_policies = collection.collection_policies
        if existing_policies:
            collection.collection_policies = {
                "public": existing_policies.get("public", {}),
                "private": dict(google_cloud_storage_project=project,
                                **existing_policies.get("private", {}))
            }
        else:
            collection.collection_policies = {
                "public": {},
                "private": {"google_cloud_storage_project": project}
            }
