import ctypes
import grp
import json
import pwd

from globus.manager.api.errors import ForbiddenError

try:
    _utils = ctypes.CDLL('libglobus_connect_gridftp_server_utils.so')
    _utils.globus_gfs_check_user_allowed_json.restype = ctypes.c_bool
    _utils.globus_gfs_check_user_allowed_json.argtypes = (
        ctypes.c_char_p,
        ctypes.c_int,
        ctypes.POINTER(ctypes.c_int),
        ctypes.c_int,
        ctypes.c_char_p)
except OSError:
    _utils = None


class PosixConnector(object):
    CONNECTOR_ID = "145812c8-decc-41f1-83cf-bb2a85a2a70b"
    DISPLAY_NAME = "POSIX"

    def __init__(self, app):
        self.default_mapping = (
            app.config['mapping']['username_without_domain'])

    def storage_credential(
            self, credential_data, valid_identity_data, storage_gateway):
        user_mapping_func = storage_gateway.storage_policies['private'].get(
            'mapping', self.default_mapping)

        username = user_mapping_func(valid_identity_data)

        if username != credential_data['storage_gateway_user_id']:
            raise ForbiddenError(
                'local identity {} does not match '
                'local account policy'.format(
                    credential_data['storage_gateway_user_id']))

        passwd_entry, groups = self.lookup_unix_info(username)
        # Forbidden if the local username is not allowed by GridFTP
        # POSIX policies
        if passwd_entry is None or not self.check_user_allowed(
                username, storage_gateway.storage_policies):
            raise ForbiddenError(
                'local identity {} does not match '
                'local account policy'.format(
                    username))

        # Return the requested username as the storage credential
        return (username, passwd_entry.pw_gecos.split(",")[0], username)

    def lookup_unix_info(self, username):
        """
        Helper method to resolve the given username to a passwd entry and
        supplement group list
        """
        supplemental_group_ids = None

        try:
            passwd_entry = pwd.getpwnam(username)
        except KeyError:
            passwd_entry = None

        if passwd_entry is not None:
            supplemental_group_ids = [
                (group.gr_gid) for group in grp.getgrall()
                if username in group.gr_mem]

        return (passwd_entry, supplemental_group_ids)

    def check_user_allowed(self, username, storage_policies):
        """
        Check whether a user is allowed to create a collection, based on a
        storage_gateway's policies.
        """
        global _utils
        passwd_entry, groups = self.lookup_unix_info(username)

        if passwd_entry is not None and _utils is not None:
            group_id = passwd_entry.pw_gid
            return _utils.globus_gfs_check_user_allowed_json(
                passwd_entry.pw_name.encode('utf8'),
                group_id,
                (ctypes.c_int * len(groups))(*groups),
                len(groups),
                json.dumps(storage_policies).encode('utf8'))
        return False
