from marshmallow import (
    validate, validates, validates_schema, post_dump, ValidationError)

from globus.manager.api.schemas import ma


class CollectionPostSchema(ma.Schema):
    # for collections that use dummy credentials (unauthenticated s3)
    # a storage gateway and identity id are given instead of a user credential
    # for all other cases a user_credential_id is required
    user_credential_id = ma.Str(allow_none=True)
    storage_gateway_id = ma.Str(allow_none=True)
    identity_id = ma.Str(allow_none=True)

    # Google Cloud Storage collections require a project
    # associated with the collection
    google_cloud_storage_project = ma.Str(allow_none=True)

    # Required definition fields
    root_path = ma.Str(missing='/')
    display_name = ma.Str(required=True, validate=[validate.Length(max=128)])

    # Optional fields
    description = ma.Str(allow_none=True, validate=[validate.Length(max=1024)])
    keywords = ma.Str(allow_none=True, validate=[validate.Length(max=1024)])
    info_link = ma.Url(allow_none=True)
    contact_email = ma.Email(allow_none=True)
    organization = ma.Str(
        allow_none=True, validate=[validate.Length(max=1024)])
    department = ma.Str(allow_none=True, validate=[validate.Length(max=1024)])
    contact_info = ma.Str(
        allow_none=True, validate=[validate.Length(max=4096)])
    default_directory = ma.Str(missing='/')
    force_encryption = ma.Boolean()

    @validates_schema
    def validate_credentials(self, data):
        ucid = data.get('user_credential_id')
        sgid = data.get('storage_gateway_id')
        iid = data.get('identity_id')
        if ((ucid and (sgid or iid))
                or (sgid and not iid)
                or (iid and not sgid)):
            raise ValidationError('Invalid credential')

    @validates('display_name')
    def validate_display_name(self, value):
        if not value.strip():
            raise ValidationError('Invalid display name')


collection_post_schema = CollectionPostSchema()


class UserCredentialFormPostSchema(ma.Schema):
    # Required fields
    identity_id = ma.Str(required=True, validate=[validate.Length(max=128)])
    redirect_uri = ma.Url(required=True)
    storage_gateway = ma.Str(required=True)
    login_hint = ma.Email(required=True)


user_credential_form_post_schema = UserCredentialFormPostSchema()


class UserCredentialJSONPostSchema(ma.Schema):
    # Required fields
    identity_id = ma.Str(required=True, validate=[validate.Length(max=128)])
    storage_gateway = ma.Str(required=True)

    # storage_gateway_user_id is only strictly required by some connectors
    storage_gateway_user_id = ma.Str(allow_none=True)

    # authenticated s3 fields
    s3_key_id = ma.Str(allow_none=True)
    s3_secret_key = ma.Str(allow_none=True)


user_credential_json_post_schema = UserCredentialJSONPostSchema()


class SharedEndpointSchema(CollectionPostSchema):
    DATA_TYPE = ma.Constant('shared_endpoint')
    host_endpoint = ma.Str(required=True)
    gcs_private_uuid = ma.Str(required=True)
    https_server = ma.Url(schemes=['https'], required=True)
    tlsftp_server = ma.Str(required=True)
    authentication_assurance_timeout = ma.Int()

    # overriden fields
    root_path = ma.Constant('/')
    storage_type = ma.Str(required=True)

    @post_dump
    def fixup(self, data):
        if data.get('user_credential_id'):
            del data['user_credential_id']
        if data.get('storage_gateway_id'):
            del data['storage_gateway_id']
        if data.get('identity_id'):
            del data['identity_id']
        if data.get('google_cloud_storage_project'):
            del data['google_cloud_storage_project']
        data['host_path'] = data['root_path']
        del data['root_path']
        if data['storage_type'] != 'Google Drive':
            data['storage_type'] = None


shared_endpoint_schema = SharedEndpointSchema()


class EndpointRoleSchema(ma.Schema):
    DATA_TYPE = ma.Constant('role')
    principal_type = ma.Str(
        default='identity',
        validate=[validate.ContainsOnly(['identity', 'group'])])
    principal = ma.Str(required=True)
    role = ma.Constant('administrator')


endpoint_role_schema = EndpointRoleSchema()
