import os
import time

from globus_sdk.base import BaseClient
from globus_sdk.exc import GlobusAPIError


class DnsClient(BaseClient):
    def __init__(self, client_id=None, authorizer=None, **kwargs):
        self.client_id = client_id
        self.default_headers = {'Content-Type': 'application/json'}
        if 'base_url' not in kwargs:
            base_url = 'https://dns.api.globus.org/'
            environment = kwargs.get('environment')
            if not environment:
                environment = os.environ.get('GLOBUS_SDK_ENVIRONMENT')
            if environment:
                if environment.lower() in [
                        'staging', 'test', 'integration', 'sandbox']:
                    base_url = 'https://dns.api.{}.globuscs.info/'.format(
                        environment.lower())
            kwargs['base_url'] = base_url
        BaseClient.__init__(
            self,
            "dns",
            base_path="/v1.0",
            authorizer=authorizer,
            **kwargs)

    def create_domain(self, a=None, aaaa=None, cname=None):
        domain = {}
        if a is not None:
            domain['a'] = a
            assert cname is None
        if aaaa is not None:
            domain['aaaa'] = aaaa
            assert cname is None
        if cname is not None:
            domain['cname'] = cname
        return self.post(
            "/domain", headers=self.default_headers, json_body=domain)

    def delete_domain(self, domain):
        return self.delete("/domain/{}".format(domain),
                           headers=self.default_headers)

    def update_domain(self, domain, a=None, aaaa=None, cname=None):
        data = {}
        if a is not None:
            assert cname is None
            data['a'] = a
            data['cname'] = []  # pass [] to remove any existing
        if aaaa is not None:
            assert cname is None
            data['aaaa'] = aaaa
            data['cname'] = []
        if cname is not None:
            assert a is None
            assert aaaa is None
            data['cname'] = cname
            data['a'] = []
            data['aaaa'] = []

        path = "/domain/{}".format(domain)
        return self.put(path, headers=self.default_headers, json_body=data)

    def create_txt(self, domain, txt, record_name="_acme-challenge"):
        data = {'txt': txt}
        path = "/domain/{0}/txt/{1}".format(domain, record_name)
        return self.put(path, headers=self.default_headers, json_body=data)

    # The following is logic for retrying on 503 ContentionErrors
    # the DNS Service returns when being rate limited by route53

    def _retry(self, f, *args, **kwargs):
        """
        Retries the given function on 503 ContentionErrors
        waits one second between retries
        """
        while True:
            try:
                return f(*args, **kwargs)
            except GlobusAPIError as e:
                if e.http_status == 503:
                    time.sleep(1)
                    continue
                else:
                    raise

    def post(self, *args, **kwargs):
        return self._retry(
            super(DnsClient, self).post, *args, **kwargs)

    def get(self, *args, **kwargs):
        return self._retry(
            super(DnsClient, self).get, *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._retry(
            super(DnsClient, self).put, *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._retry(
            super(DnsClient, self).delete, *args, **kwargs)
