from flask import current_app
from cachelib import SimpleCache

from globus.manager.api.utils.services import TRANSFER_SCOPE
import globus_sdk

CLIENTS_API_BASE_URL = '/v2/api/clients'

cache = SimpleCache()


def _virtual_client_representation(collection, migration_uuid=None):
    vc = {
        'client': {
            'name': collection.domain,
            'public_client': False,
            'visibility': 'public',
            'redirect_uris': [
                'https://{}/redirect'.format(collection.domain)
            ]
        }
    }
    if migration_uuid is not None:
        vc['client']['google_drive_collection_id'] = migration_uuid

    return vc


def _scope_name_to_id(scope_name):
    cache_key = 'transfer_scope_id'

    scope_id = cache.get(cache_key)
    if scope_id:
        return scope_id

    owner_client_id = current_app.config['CLIENT_ID']
    owner_secret = current_app.config['CLIENT_SECRET']

    authorizer = globus_sdk.BasicAuthorizer(owner_client_id, owner_secret)
    auth_client = globus_sdk.AuthClient(
        client_id=owner_client_id, authorizer=authorizer)

    path = '/v2/api/scopes?scope_strings={}'.format(scope_name)

    res = auth_client.get(path)

    scope_id = res.data['scopes'][0]['id']
    cache.set(cache_key, scope_id, 7200)

    return scope_id


def _scope_representation(collection, collection_client_id):
    return {
        'scope': {
            'name': collection.domain,
            'description': 'HTTPS for {}'.format(collection.domain),
            'scope_suffix': 'https',
            'client': collection.id,
            'dependent_scopes': [
                {
                    'scope': _scope_name_to_id(TRANSFER_SCOPE),
                    'optional': False,
                    'requires_refresh_token': False,
                    'allows_refresh_token': True
                }
            ]
        }
    }


def register_virtual_client(collection, migration_uuid=None):
    owner_client_id = current_app.config['CLIENT_ID']
    owner_secret = current_app.config['CLIENT_SECRET']

    authorizer = globus_sdk.BasicAuthorizer(owner_client_id, owner_secret)
    auth_client = globus_sdk.AuthClient(
        client_id=owner_client_id, authorizer=authorizer)

    if migration_uuid is not None:
        try:
            res = auth_client.get(
                '{}/{}'.format(
                    CLIENTS_API_BASE_URL,
                    migration_uuid))
            if res['client']['parent_client'] == owner_client_id:
                return res.data
        except globus_sdk.exc.AuthAPIError:
            pass

    res = auth_client.post(
        CLIENTS_API_BASE_URL, _virtual_client_representation(
            collection, migration_uuid=migration_uuid))

    return res.data


def register_scope(collection, collection_client_id):
    owner_secret = current_app.config['CLIENT_SECRET']

    authorizer = globus_sdk.BasicAuthorizer(collection_client_id, owner_secret)
    auth_client = globus_sdk.AuthClient(
        client_id=collection_client_id, authorizer=authorizer)

    res = auth_client.post(
        '{}/{}/scopes'.format(CLIENTS_API_BASE_URL, collection_client_id),
        _scope_representation(collection, collection_client_id))

    return res.data


def register_fqdn(collection, collection_client_id):
    owner_secret = current_app.config['CLIENT_SECRET']

    authorizer = globus_sdk.BasicAuthorizer(collection_client_id, owner_secret)
    auth_client = globus_sdk.AuthClient(
        client_id=collection_client_id, authorizer=authorizer)

    res = auth_client.post(
        '{}/{}/fqdns'.format(CLIENTS_API_BASE_URL, collection_client_id),
        {'fqdn': collection.domain})

    return res.data


def register(collection, migration_uuid=None):
    client = register_virtual_client(collection, migration_uuid=migration_uuid)
    collection.id = client['client']['id']
    fqdn = register_fqdn(collection, collection.id)
    scope = register_scope(collection, collection.id)

    return client['client'], scope['scopes'], fqdn['fqdn']
