import argparse
import copy
import os
import pwd
import stat
import subprocess
import sys


def main(argv=sys.argv[1:]):
    if os.geteuid() == 0:
        WORK_DIR_DEFAULT = '/var/lib/letsencrypt'
        LOGS_DIR_DEFAULT = '/var/log/letsencrypt'
        CONFIG_DIR_DEFAULT = '/etc/letsencrypt'
    else:
        homedir = pwd.getpwuid(os.geteuid()).pw_dir
        WORK_DIR_DEFAULT = os.path.join(homedir, 'lib/letsencrypt')
        LOGS_DIR_DEFAULT = '/var/log/globus-connect-server/letsencrypt'
        CONFIG_DIR_DEFAULT = os.path.join(homedir, 'etc/letsencrypt')

    parser = argparse.ArgumentParser()
    parser.add_argument('--domain', help='domain to get a cert for')
    parser.add_argument(
        '--work-dir',
        help='certbot work dir',
        default=WORK_DIR_DEFAULT)
    parser.add_argument(
        '--logs-dir',
        help='certbot logs dir',
        default=LOGS_DIR_DEFAULT)
    parser.add_argument(
        '--config-dir',
        help='certbot config dir',
        default=CONFIG_DIR_DEFAULT)
    parser.add_argument(
        '--challenge',
        choices=['dns', 'http'],
        help='ACME challenge type',
        default='dns')
    parser.add_argument(
        '--staging', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument(
        '--force', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument(
        '--renew', action='store_true', help='Renew certificates')
    parser.add_argument(
        '--quiet',
        action='store_true',
        help='Enable quiet mode')

    args = parser.parse_args(args=argv)
    os.umask(0o77)

    if not args.renew and args.domain is None:
        print("Missing required argument --domain")
        sys.exit(1)
    if args.renew:
        cmd = ['certbot', 'renew']
        if args.force:
            cmd.append('--force-renewal')
    else:
        cmd = ['certbot', 'certonly', '-d', args.domain, '--keep']

    cmd.extend([
        '--register-unsafely-without-email',
        '--preferred-challenges',
        args.challenge,
        '--manual-public-ip-logging-ok',
        '--agree-tos',
        '--work-dir', args.work_dir,
        '--config-dir', args.config_dir,
        '--logs-dir', args.logs_dir,
    ])
    if args.staging:
        cmd.append('--staging')
    if args.challenge == 'http':
        cmd.extend(['--webroot', '--webroot-path', '/var/www/html'])
    elif args.challenge == 'dns':
        cmd.extend([
            '--manual',
            '--manual-auth-hook',
            '/opt/globus/bin/certbot_dns_challenge_auth_hook'
        ])
    else:
        cmd.extend(['--manual', '--manual-auth-hook', 'true'])

    if not os.path.exists(args.work_dir):
        os.makedirs(args.work_dir)
    if not os.path.exists(args.config_dir):
        os.makedirs(args.config_dir)
    if not os.path.exists(args.logs_dir):
        os.makedirs(args.logs_dir)

    cmdenv = copy.deepcopy(os.environ)
    proc = subprocess.Popen(cmd, env=cmdenv,
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()
    if proc.returncode != 0:
        if not args.quiet:
            print(out.decode('utf8'))
            print(err.decode('utf8'), file=sys.stderr)
            raise Exception("certbot error {}".format(proc.returncode))
        else:
            print(out)
            print(err, file=sys.stderr)
    elif args.domain is not None:
        cert = '{0}/live/{1}/cert.pem'.format(
            args.config_dir,
            args.domain)
        os.chmod(cert,
                 stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
        key = '{0}/live/{1}/privkey.pem'.format(
            args.config_dir,
            args.domain)
        os.chmod(key, stat.S_IRUSR | stat.S_IWUSR)
        chain = '{0}/live/{1}/fullchain.pem'.format(
            args.config_dir,
            args.domain)
        if not args.quiet:
            print('cert={0}'.format(cert))
            print('key={0}'.format(key))
            print('chain={0}'.format(chain))
    return proc.returncode


if __name__ == '__main__':
    sys.exit(main())
