import argparse
import json
import sys

from sqlalchemy.orm.exc import NoResultFound

from globus.manager.api import make_app
from globus.manager.api.models import (
    db, Configuration, Connector, StorageGateway)
from globus.manager.api.schemas import ma


app = make_app()
db.init_app(app)
ma.init_app(app)


def fetch_connector(connector_id):
    try:
        return Connector.query.filter_by(id=connector_id).one()
    except NoResultFound:
        raise NoResultFound('{} is not a valid connector_id'.format(
            connector_id))


def fetch_or_create_storage_gateway(connector, display_name):
    try:
        ss = StorageGateway.query.filter_by(
            display_name=display_name,
            connector_id=connector.id).one()
    except NoResultFound:
        sg = StorageGateway()
        sg.display_name = display_name
        sg.connector = connector
        db.session.add(sg)

    return ss


def update_storage_credential(connector_id, display_name, data):
    with app.app_context():
        connector = fetch_connector(connector_id)
        ss = fetch_or_create_storage_gateway(connector, display_name)
        ss.credential = data

        db.session.commit()


def update_storage_policies(connector_id, display_name, data):
    with app.app_context():
        connector = fetch_connector(connector_id)
        ss = fetch_or_create_storage_gateway(connector, display_name)
        ss.storage_policies = data

        db.session.commit()


def add_or_update_configuration(name, value):
    with app.app_context():
        try:
            config = Configuration.query.filter_by(name=name).one()
            config.value = value
        except NoResultFound:
            config = Configuration(name, value)
            db.session.add(config)

        db.session.commit()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'destination',
        help='configuration, storage_credential, or storage_policies.')
    parser.add_argument(
        'identifier',
        help='the name of the configuration item or the display name of '
        'the storage gateway.')
    parser.add_argument('data')
    parser.add_argument(
        '--connector_id',
        help='required for storage_credential and storage_policies.')
    args = parser.parse_args()

    if args.data == 'stdin':
        args.data = sys.stdin.read()

    if args.destination.lower() == 'configuration':
        add_or_update_configuration(args.identifier, args.data)
    elif args.destination.lower() == 'storage_credential':
        if not args.connector_id:
            parser.error(
                '--connector_id required for storage_credential')
        update_storage_credential(
            args.connector_id, args.identifier, json.loads(args.data))
    elif args.destination.lower() == 'storage_policies':
        if not args.connector_id:
            parser.error(
                '--connector_id required for storage_policies')
        update_storage_policies(
            args.connector_id, args.identifier, json.loads(args.data))


if __name__ == '__main__':
    main()
