import os
import click
from sqlalchemy.orm.exc import NoResultFound
from ..app import app, db
from globus.manager.api.utils.services import get_transfer_client
from globus.manager.api.utils.https_endpoint import HTTPSEndpoint
from globus.manager.api.models import (
    Collection,
    Configuration,
    PreallocatedCollection,
    StorageGateway,
    UserCredential,
)
from globus.manager.version import __version__
import sqlalchemy


def create_gridmap(collection_id):
    gridmap_path = (
        '/var/lib/globus-connect-server/grid-security/gridmap.{}'.format(
            collection_id))

    if not os.path.exists(gridmap_path):
        gridmap_dir = os.path.dirname(gridmap_path)
        if not os.path.exists(gridmap_dir):
            os.makedirs(os.path.dirname(gridmap_path), 0o755)
        old_umask = os.umask(0o22)
        open(gridmap_path, 'w').close()
        os.umask(old_umask)

    return gridmap_path


@click.command(
    name='create',
    short_help='Create a collection',
    help='''\
The collection create subcommand creates a new collection. Currently
this supports creating "mapped" collections---those which map an OAuth
identity username to a local user account. Visit https://globus.org to create
"guest" collections.

The storage gateway id (returned
from the `globus-connect-server-config storage gateway create` or
`globus-connect-server-config storage gateway list` commands) and the
display name of the collection are required to create a new collection.

The newly created collection inherits its policy from the storage
gateway it is associated with.

== EXAMPLES

Create a POSIX storage gateway and then a mapped collection rooted
at the storage gateway.
....
    $ sudo globus-connect-server-config storage-gateway create \\
        -r '$HOME' -t "POSIX" -n "Home directories" -d uchicago.edu \\
        --allow-mapped-collections --disallow-guest-collections \\
    Storage Gateway Created: f97ca3dc-96b2-4923-97ac-3291ffefe74f

    $ sudo globus-connect-server-config collection create \\
        -i f97ca3dc-96b2-4923-97ac-3291ffefe74f \\
        -n "Mapped Home Directories"
    Created collection 1e16f4c0-a7cd-4c58-9e6a-3684d92a3de9
....
''')
@click.help_option('--help', '-h')
@click.version_option(
    version=__version__,
    prog_name='globus-connect-server-setup')
@click.option(
    '--storage-gateway-id', '-i',
    metavar='ID',
    required=True,
    help='Create a collection on storage gateway ID')
@click.option(
    '--collection-type', '-t',
    type=click.Choice(['mapped']),
    help='Type of collection to create',
    default='mapped')
@click.option(
    '--display-name', '-n',
    metavar='NAME',
    required=True,
    help='Display name of the collection')
@click.option(
    '--root', '-r',
    metavar='ROOT',
    help='Root directory of the collection.  The default is /, \
         which is the storage gateway root',
    default='/')
@click.option(
    '--default-directory', '-d',
    metavar='DIRECTORY',
    help='Default directory when browsing the collection')
@click.pass_context
def collection_create(
        ctx, storage_gateway_id, collection_type, display_name, root,
        default_directory):
    GLOBUS_MAPPING_ID = ':globus-mapping:'
    with app.app_context():
        sg = None
        uc = None
        try:
            sg = StorageGateway.query.filter_by(id=storage_gateway_id).one()
            uc = UserCredential.query.filter(
                UserCredential.storage_gateway_id == sg.id,
                UserCredential.identity_id == GLOBUS_MAPPING_ID).one_or_none()
            required_domains = (
                sg.storage_policies.get('public', {}).get('domains', []))
            if uc is None:
                uc = UserCredential(
                    display_name='Dummy credential for mapped collections',
                    identity_id=GLOBUS_MAPPING_ID,
                    storage_gateway_user_id=GLOBUS_MAPPING_ID,
                    storage_gateway_id=sg.id)
                db.session.add(uc)
                db.session.commit()
        except NoResultFound:
            click.echo('No such storage gateway')
            ctx.exit(1)
        except sqlalchemy.exc.OperationalError:
            click.echo('Permission denied')
            ctx.exit(1)

        if not sg.allow_mapped_collections:
            click.echo(
                "Storage gateway {} does not allow mapped collections."
                .format(sg.display_name))
            ctx.exit(1)
        elif not sg.connector.allow_mapped_collections:
            click.echo("Connector does not allow mapped collections.")
            ctx.exit(1)
        elif not sg.require_high_assurance:
            click.echo(
                "Storage gateway {} is not high assurance."
                .format(sg.display_name))
            ctx.exit(1)

        client_id = Configuration.query.filter_by(
            name='client_id').one().value
        secret = Configuration.query.filter_by(
            name='client_secret').one().value
        environment = Configuration.query.filter_by(
            name='environment').one().value

        pc = PreallocatedCollection.query.order_by(
            PreallocatedCollection.created_at).first()
        if pc is None:
            click.echo('No collections available. Try again later.')
            ctx.exit(1)
        db.session.delete(pc)
        db.session.commit()

        gridmap = create_gridmap(pc.id)
        policies = {
            'public': {
                'type': 'mapped'
            },
            'private': {
                'gridmap': gridmap
            }
        }
        if default_directory is not None:
            policies['private']['default_directory'] = default_directory

        c = Collection(
            id=pc.id,
            path=root,
            domain=pc.domain,
            collection_policies=policies,
            user_credential_id=uc.id)
        db.session.add(c)
        db.session.commit()

        transfer_client = get_transfer_client()
        try:
            host_endpoint_id = Configuration.query.filter_by(
                name='host_endpoint_id').one().value
            ep_data = {
                'DATA_TYPE': 'endpoint',
                'display_name': display_name,
                'public': True,
                'myproxy_server': 'myproxy.globusonline.org',
                'storage_type': sg.connector.display_name,
                'default_directory': '/{server_default}/',
                'gcs_private_uuid': pc.id,
                'tlsftp_server': 'tlsftp://{}:443'.format(pc.domain),
                'https_server': 'https://{}'.format(pc.domain),
                'gcs_private_parent_endpoint_id': host_endpoint_id,
            }

            collection_flags = ''
            authentication_assurance_timeout = None
            public_policies = sg.storage_policies['public']
            if sg.require_high_assurance:
                ep_data['force_encryption'] = True
                ep_data['authentication_assurance_timeout'] = int(
                    public_policies['authentication_assurance_timeout']) * 60
                collection_flags = "RequireHighAssurance"
                authentication_assurance_timeout = (
                    public_policies['authentication_assurance_timeout'])

            transfer_client.create_endpoint(ep_data)

            if len(required_domains) > 1:
                click.echo(
                    'Warning: multiple required domains defined, '
                    'only using {}'.format(required_domains[0]))
                required_domain = required_domains[0]
            elif len(required_domains) > 0:
                required_domain = required_domains[0]

            https_endpoint = HTTPSEndpoint(
                pc.id, client_id, secret,
                pc.domain, collection_type, required_domain,
                collection_flags=collection_flags,
                environment=environment.lower(),
                authentication_assurance_timeout=(
                    authentication_assurance_timeout))
            https_endpoint.setup()

            click.echo('Created collection {}'.format(pc.id))
        except Exception as e:
            click.echo(str(e))
            db.session.delete(c)
            db.session.commit()


if __name__ == '__main__':
    collection_create()
