import os
import click
from sqlalchemy.orm.exc import NoResultFound
from ..app import app, db
from globus.manager.api.models import Collection, Configuration
from globus.manager.api.utils.services import (
    get_dns_client, get_transfer_client)
from globus.manager.version import __version__
import globus_sdk
import sqlalchemy


@click.command(
    name='delete',
    short_help='Delete a collection',
    help='''\
The collection delete subcommand deletes a collection. This permanently
removes the collection, including the globus.org Endpoint, the GridFTP server
configuration, and the HTTPS interface to the collection.''')
@click.help_option('--help', '-h')
@click.version_option(
    version=__version__,
    prog_name='globus-connect-server-setup')
@click.argument('collection-id')
@click.option('--force', '-f', is_flag=True, help="Don't prompt for deletion")
@click.pass_context
def collection_delete(ctx, collection_id, force, no_commit=False):
    with app.app_context():
        collections = []
        try:
            if collection_id == 'all':
                collections = Collection.query.all()
            else:
                collections = [
                    Collection.query.filter_by(id=collection_id).one()
                ]
        except NoResultFound:
            click.echo('No such collection')
        except sqlalchemy.exc.OperationalError:
            click.echo('Permission denied')
            ctx.exit(1)

        for c in collections:
            prompt = 'Delete collection {}'.format(c.id)
            if force or click.confirm(prompt):
                # delete collection definition in transfer
                transfer_client = get_transfer_client()
                try:
                    transfer_client.delete_endpoint(c.id)
                except globus_sdk.exc.TransferAPIError as e:
                    if e.http_status == 404:
                        click.echo(str(e).strip())
                    else:
                        raise(e)

                dns_client = get_dns_client()
                try:
                    dns_client.delete_domain(c.domain)
                except globus_sdk.exc.GlobusAPIError as e:
                    if e.http_status == 404:
                        click.echo(str(e).strip())
                    else:
                        raise(e)

                # delete collection's conf, cert renewal, and gridmap files
                https_collection_path = os.path.expanduser(
                    '~gcsweb/etc/httpd/conf.d/{0}.conf'.format(c.id))
                if os.path.exists(https_collection_path):
                    os.remove(https_collection_path)

                letsencrypt_live_path = os.path.expanduser(
                    '~gcsweb/etc/live/{0}'.format(c.domain))
                letsencrypt_archive_path = os.path.expanduser(
                    '~gcsweb/etc/archive/{0}'.format(c.domain))
                letsencrypt_renewal_path = os.path.expanduser(
                    '~gcsweb/etc/letsencrypt/renewal/{0}.conf'.format(
                        c.domain))
                if os.path.exists(letsencrypt_live_path):
                    for p in os.listdir(letsencrypt_live_path):
                        full = os.path.join(letsencrypt_live_path, p)
                        if os.path.exists(full):
                            os.remove(full)
                    os.rmdir(letsencrypt_live_path)
                if os.path.exists(letsencrypt_archive_path):
                    for p in os.listdir(letsencrypt_archive_path):
                        full = os.path.join(letsencrypt_archive_path, p)
                        if os.path.exists(full):
                            os.remove(full)
                    os.rmdir(letsencrypt_archive_path)
                if os.path.exists(letsencrypt_renewal_path):
                    os.remove(letsencrypt_renewal_path)

                gridmap_path = (
                    '/var/lib/globus-connect-server/grid-security/gridmap.{}'
                    .format(c.id))
                if os.path.exists(gridmap_path):
                    os.remove(gridmap_path)

                # delete the collection's auth client
                try:
                    client_id = Configuration.query.filter_by(
                        name='client_id').one().value
                    client_secret = Configuration.query.filter_by(
                        name='client_secret').one().value
                    auth_client = globus_sdk.AuthClient(
                        client_id=client_id,
                        authorizer=globus_sdk.BasicAuthorizer(
                            client_id, client_secret))
                    auth_client.delete('/v2/api/clients/{0}'.format(
                        c.id))
                except globus_sdk.exc.AuthAPIError as e:
                    if e.http_status == 404:
                        click.echo(str(e).strip())
                    else:
                        raise(e)

                # finally delete the collection from the db
                db.session.delete(c)
                if not no_commit:
                    db.session.commit()


if __name__ == '__main__':
    collection_delete()
