import click

import globus_sdk
from globus.manager.api.utils.services import get_transfer_client
from globus.manager.api.models import (
    Collection,
    Connector,
    StorageGateway,
    UserCredential
)
from globus.manager.version import __version__
from ..app import app
from ..storage_gateway.constants import CONNECTOR_DISPLAY_NAMES
import sqlalchemy
import sqlalchemy.orm.exc


def display_collection(collection, display_name, collection_type, displayed):

    # if this is the first collection to be displayed, show the header
    if not displayed:
        click.echo('ID | Connector | Storage Gateway | Type | Display Name | Domain') # noqa

    storage_gateway = StorageGateway.query.join(UserCredential).filter(
        UserCredential.id == collection.user_credential_id).one()

    click.echo(
        '{0} | {1} | {2} | {3} | {4} | {5}'.format(
            collection.id,
            storage_gateway.connector.display_name,
            storage_gateway.display_name,
            collection_type,
            display_name,
            collection.domain))


@click.command(
    name='list',
    short_help='List collections',
    help='''\
The collection list sub-command genrates a list of the collections that exist
on a Globus Connect Server. The list may be filtered by a
combination of the the connector, storage gateway id, collection type,
display name, or collection id.''')
@click.help_option('--help', '-h')
@click.version_option(
    version=__version__,
    prog_name='globus-connect-server-setup')
@click.argument('COLLECTION_ID', required=False)
@click.option(
    '--connector', '-c',
    type=click.Choice(CONNECTOR_DISPLAY_NAMES),
    help='List collections on connectors of TYPE')
@click.option(
    '--storage-gateway-id', '-i',
    metavar='ID',
    help='List collections on this storage gateway')
@click.option(
    '--display-name', '-n',
    metavar='NAME',
    help='List collections with this NAME')
@click.option(
    '--collection-type', '-t',
    metavar='TYPE',
    type=click.Choice(['guest', 'mapped']),
    help='List collections with this TYPE')
@click.pass_context
def collection_list(
        ctx, collection_id, connector, storage_gateway_id, display_name,
        collection_type):
    with app.app_context():
        try:
            transfer_client = get_transfer_client()
        except sqlalchemy.exc.OperationalError:
            click.echo('Permission denied')
            ctx.exit(1)
        except sqlalchemy.orm.exc.NoResultFound:
            click.echo('Endpoint not configured')
            ctx.exit(1)

        displayed = False
        query = Collection.query.join(
            UserCredential).join(StorageGateway).join(Connector)

        # query filters
        if connector:
            query = query.filter(sqlalchemy.func.lower(Connector.display_name)
                                 == connector.lower())
        if storage_gateway_id:
            query = query.filter(StorageGateway.id == storage_gateway_id)

        # query
        if collection_id:
            results = [query.filter(Collection.id == collection_id).one()]
        else:
            results = query.all()

        # post query filters and display
        for collection in results:

            try:
                this_collection_type = (
                    collection.collection_policies['public']['type'])
            except (ValueError, KeyError):
                this_collection_type = '[MALFORMED]'
            if collection_type and this_collection_type != collection_type:
                continue

            try:
                endpoint = transfer_client.get_endpoint(collection.id)
                endpoint_display_name = endpoint.data['display_name']
            except globus_sdk.exc.TransferAPIError:
                endpoint_display_name = '[deleted]'
            if display_name and endpoint_display_name != display_name:
                continue

            display_collection(collection, endpoint_display_name,
                               this_collection_type, displayed)
            displayed = True

        if not displayed:
            click.echo('No matching collections found')


if __name__ == '__main__':
    collection_list()
