import os

import click
import globus_sdk

from ...app import app
from globus.manager.api.models import Configuration, Collection
from globus.manager.api.utils.services import get_transfer_client
from globus.manager.api.schemas.request import endpoint_role_schema
from globus.manager.version import __version__
import sqlalchemy.exc

VIEW_IDENTITIES_SCOPE = 'urn:globus:auth:scope:auth.globus.org:view_identities'


@click.command(
    'add-role',
    short_help='Add a new administrator to a Globus Connect Server endpoint',
    help='Add a new administrator to a Globus Connect Server endpoint')
@click.argument('identity', required=False)
@click.help_option('--help', '-h')
@click.version_option(
    version=__version__,
    prog_name='globus-connect-server-config')
@click.pass_context
def endpoint_admin_add_role(ctx, identity):
    add_role_failed = 0
    with app.app_context():
        try:
            globus_client_id = Configuration.query.filter_by(
                name='client_id').one().value
            globus_client_secret = Configuration.query.filter_by(
                name='client_secret').one().value
            host_endpoint_id = Configuration.query.filter_by(
                name='host_endpoint_id').one().value
            globus_sdk_environment = Configuration.query.filter_by(
                name='environment').one().value.lower()
            os.environ['GLOBUS_SDK_ENVIRONMENT'] = globus_sdk_environment

            if identity is None:
                identity = click.prompt("Identity")

            client = globus_sdk.ConfidentialAppAuthClient(
                globus_client_id, globus_client_secret)
            tokens = client.oauth2_client_credentials_tokens(
                requested_scopes=VIEW_IDENTITIES_SCOPE).by_resource_server
            auth_token = tokens['auth.globus.org']['access_token']
            auth_client = globus_sdk.AuthClient(
                authorizer=globus_sdk.AccessTokenAuthorizer(auth_token))

            r = auth_client.get_identities(usernames=[identity])

            identities = r['identities']
            if len(identities) == 0:
                click.echo('Unknown globus identity {}.'.format(identity))
                ctx.exit(1)
            else:
                role_data, _ = endpoint_role_schema.dump(
                    {'principal': r['identities'][0]['id']})

            transfer_client = get_transfer_client()
            try:
                transfer_client.add_endpoint_role(host_endpoint_id, role_data)
                click.echo(
                    'Added an administrator role for {} to '
                    'endpoint {}'.format(
                        identity, host_endpoint_id))
            except globus_sdk.exc.TransferAPIError as e:
                click.echo(
                    'Failed to add administrator role for {} to '
                    'endpoint {}: {}'.format(
                        identity, host_endpoint_id, str(e)))
                add_role_failed += 1

            for c in Collection.query.all():
                if c.collection_policies['public']['type'] == 'mapped':
                    try:
                        transfer_client.add_endpoint_role(c.id, role_data)
                        click.echo(
                            'Added an administrator role for {} to '
                            'endpoint {}'.format(
                                identity, c.id))
                    except globus_sdk.exc.TransferAPIError as e:
                        click.echo(
                            'Failed to add administrator role for {} to '
                            'endpoint {}: {}'.format(
                                identity, host_endpoint_id, str(e)))
                        add_role_failed += 1
        except globus_sdk.exc.TransferAPIError as e:
            click.echo(str(e))
            ctx.exit(1)
        except sqlalchemy.exc.OperationalError:
            click.echo('Permission denied')
            ctx.exit(1)
        ctx.exit(add_role_failed)
