from __future__ import absolute_import

import click
from sqlalchemy.exc import (IntegrityError, OperationalError)
from ..app import app, db
from ..utils import BadInvocation
from globus.manager.api.models import (
    Connector,
    StorageGateway
)
from globus.manager.version import __version__
from .constants import CONNECTOR_DISPLAY_NAMES
from .update import (
    add_options,
    common_options,
    google_options,
    google_drive_options,
    posix_options,
    s3_ceph_and_blackpearl_options,
    s3_and_ceph_options,
    s3_only_options,
    ceph_only_options,
    blackpearl_only_options,
    box_options,
    google_cloud_storage_options,
    gateway_subscription_check,
    storage_gateway_update_storage,
    SHARED_OPTION_HELP_TEXT,
)


@click.command(
    'create',
    short_help='Create a storage gateway',
    help='''\
The storage-gateway create subcommand creates a new storage gateway, which
is a named location for creating collections. It also can associate optional
policies governing path restrictions and identity providers which may access
the storage gateway. Once a storage gateway is created, visit globus.org to
create guest collections to access the storage gateway.

The --root, --connector, and --display-name parameters are required to
create a storage gateway.

The --high-assurance flag and --authentication-assurance-timeout option are
required to create a high assurance storage gateway.

{}'''.format(SHARED_OPTION_HELP_TEXT))
@click.help_option('--help', '-h')
@click.version_option(
    version=__version__,
    prog_name='globus-connect-server-config')
@click.option('--root', '-r', metavar='ROOT', help='Storage gateway root path')
@click.option('--high-assurance', is_flag=True,
              help='Enable high assurance features')
@click.option('--connector', '-t', help='Connector type',
              type=click.Choice(CONNECTOR_DISPLAY_NAMES))
@add_options(
    [(name, opt) for (name, opt) in common_options if name != 'domain'])
@add_options(posix_options)
@add_options(google_options)
@add_options(google_drive_options)
@add_options(s3_ceph_and_blackpearl_options)
@add_options(s3_and_ceph_options)
@add_options(s3_only_options)
@add_options(ceph_only_options)
@add_options(blackpearl_only_options)
@add_options(box_options)
@add_options(google_cloud_storage_options)
@click.option('--domain', '-d', metavar='DOMAIN', multiple=True, required=True,
              help='Allowed username domain')
@click.pass_context
def storage_gateway_create(ctx, **kwargs):
    for r in ['root', 'display_name', 'connector']:
        if r not in kwargs or kwargs.get(r) is None:
            kwargs[r] = click.prompt('Storage System {}'.format(
                r.replace('_', ' ')))
    if kwargs['connector'] == 'Google Drive':
        for r in ['client_id', 'client_secret']:
            if r not in kwargs or kwargs.get(r) is None:
                kwargs[r] = click.prompt('Google Drive {}'.format(
                    r.replace('_', ' ')))

    with app.app_context():
        try:
            if (kwargs.get('authentication_assurance_timeout')
                    and not kwargs.get('high_assurance')):
                raise BadInvocation(
                    'The --high-assurance flag is required to '
                    'set the authentication assurance timeout')
            elif (kwargs.get('high_assurance')
                  and not kwargs.get('authentication_assurance_timeout')):
                raise BadInvocation(
                    'The --high-assurance flag requires the '
                    '--authentication-assurance-timeout option')
            connector = Connector.query.filter_by(
                display_name=kwargs['connector']).one()

            # check if this action requires a subscription
            gateway_subscription_check(**kwargs)

            # set values from create only options and init storage policies
            sg = StorageGateway()
            sg.connector = connector
            private_policies = {}
            public_policies = {'root': kwargs['root']}
            sg.storage_policies = {
                'public': public_policies,
                'private': private_policies,
            }

            # set values from shared options and commit to db
            storage_gateway_update_storage(sg, **kwargs)

            click.echo('Storage Gateway Created: {0}'.format(sg.id))
        except IntegrityError:
            db.session.rollback()
            nsg = StorageGateway.query.filter_by(
                display_name=kwargs['display_name']).one()
            click.echo('Storage Gateway "{0}" already exists: {1}'.format(
                nsg.display_name,
                nsg.id))
            ctx.exit(1)
        except OperationalError:
            click.echo('Permission denied')
            ctx.exit(1)
        except BadInvocation as e:
            click.echo(str(e))
            click.get_current_context().exit(1)


if __name__ == '__main__':
    storage_gateway_create()
