import json
import click

from ..app import app
from globus.manager.api.models import (StorageGateway)
from globus.manager.version import __version__
from sqlalchemy.exc import OperationalError


def storage_gateway_show_common_connector(sg):
    idps = sg.storage_policies['public'].get('identity_providers')
    if idps is not None:
        for idp in idps:
            click.echo('--identity-provider {}'.format(idp))

    domains = sg.storage_policies['public'].get('domains')
    if domains is not None:
        for domain in domains:
            click.echo('--domain "{}"'.format(domain))
    click.echo('--connector "{}"'.format(sg.connector.display_name))


def storage_gateway_show_common_storage_gateway(sg):
    click.echo('--display-name "{}"'.format(sg.display_name))

    rp = sg.storage_policies['private'].get('restrict_paths')
    if rp:
        click.echo('--restrict-paths "{}"'.format(rp))

    st = sg.storage_policies['public'].get('authentication_assurance_timeout')
    if st:
        click.echo('--authentication-assurance-timeout {}'.format(st))

    if sg.allow_guest_collections:
        click.echo('--allow-guest-collections')
    else:
        click.echo('--disallow-guest-collections')

    if sg.allow_mapped_collections:
        click.echo('--allow-mapped-collections')
    else:
        click.echo('--disallow-mapped-collections')

    if sg.require_high_assurance:
        click.echo('--high-assurance')

    click.echo('--root "{}"'.format(
        sg.storage_policies['public'].get('root')))


def storage_gateway_show_posix(sg):
    click.secho(sg.id, bold=True)

    storage_gateway_show_common_connector(sg)
    storage_gateway_show_common_storage_gateway(sg)

    allowdeny = ['users_allow', 'users_deny', 'groups_allow', 'groups_deny']

    for key in allowdeny:
        val = sg.storage_policies['private'].get(key)
        if val is not None:
            click.echo('--{0} {1}'.format(key.replace('_', '-'), val))


def storage_gateway_show_google_drive(sg):
    click.secho(sg.id, bold=True)

    client_id = sg.credential['client_id']

    click.echo('--client-id "{}"'.format(client_id))

    client_secret = sg.credential['secret']
    click.echo('--client-secret "{}"'.format(client_secret))

    storage_gateway_show_common_connector(sg)

    load_dsi_module = sg.storage_policies['private'].get('load_dsi_module')
    quota = None
    for opt in load_dsi_module.split(':'):
        if opt.startswith('api_quota_rate='):
            quota = int(opt.replace('api_quota_rate=', ''))
    if quota is not None:
        click.echo('--user-api-rate-quota {0}'.format(quota))

    storage_gateway_show_common_storage_gateway(sg)


def storage_gateway_show_s3_and_ceph(sg):
    dsi_config = sg.storage_policies['private'].get('dsi_config', {})

    s3_endpoint = dsi_config.get('s3_endpoint')
    click.echo('--s3-endpoint "{}"'.format(s3_endpoint))

    for bucket in dsi_config.get('s3_buckets'):
        click.echo('--s3-bucket "{}"'.format(bucket))


def storage_gateway_show_s3(sg):
    click.secho(sg.id, bold=True)

    storage_gateway_show_common_connector(sg)
    storage_gateway_show_common_storage_gateway(sg)
    storage_gateway_show_s3_and_ceph(sg)

    dsi_config = sg.storage_policies['private'].get('dsi_config', {})
    s3_user_credential = dsi_config.get('s3_user_credential_available')
    if s3_user_credential:
        click.echo('--s3-user-credential')
    else:
        click.echo('--s3-unauthenticated')


def storage_gateway_show_ceph(sg):
    click.secho(sg.id, bold=True)

    storage_gateway_show_common_connector(sg)
    storage_gateway_show_common_storage_gateway(sg)
    storage_gateway_show_s3_and_ceph(sg)

    dsi_config = sg.storage_policies['private'].get('dsi_config', {})
    for key in ['ceph_admin_key_id', 'ceph_admin_secret_key']:
        val = dsi_config.get(key)
        if val is not None:
            click.echo('--{0} "{1}"'.format(key.replace('_', '-'), val))


def storage_gateway_show_blackpearl(sg):
    click.secho(sg.id, bold=True)

    storage_gateway_show_common_connector(sg)
    storage_gateway_show_common_storage_gateway(sg)

    dsi_config = sg.storage_policies['private'].get('dsi_config', {})
    s3_endpoint = dsi_config.get('s3_endpoint')
    click.echo('--s3-endpoint "{}"'.format(s3_endpoint))

    for key in ['bp_access_id_file']:
        val = dsi_config.get(key)
        if val is not None:
            click.echo('--{0} "{1}"'.format(key.replace('_', '-'), val))


def storage_gateway_show_box(sg):
    click.secho(sg.id, bold=True)

    storage_gateway_show_common_connector(sg)
    storage_gateway_show_common_storage_gateway(sg)
    dsi_config = sg.storage_policies['private'].get('dsi_config', {})

    # since dsi_config is set to a passed JSON file, we just return
    # the full dsi_config as the original value of that file
    click.echo("Box JSON config:")
    click.echo(json.dumps(
        dsi_config, indent=4, sort_keys=True, separators=(',', ': ')))


def storage_gateway_show_google_cloud_storage(sg):
    click.secho(sg.id, bold=True)

    storage_gateway_show_common_connector(sg)
    storage_gateway_show_common_storage_gateway(sg)

    click.echo('--client-id "{}"'.format(sg.credential['client_id']))
    click.echo('--client-secret "{}"'.format(sg.credential['secret']))

    dsi_config = sg.storage_policies['private'].get('dsi_config', {})

    for bucket in dsi_config.get('google_cloud_storage_buckets'):
        click.echo('--google-cloud-storage-bucket "{}"'.format(bucket))
    for project in dsi_config.get('google_cloud_storage_projects'):
        click.echo('--google-cloud-storage-project "{}"'.format(project))


@click.command(name='show', help='Show storage gateway details')
@click.help_option('--help', '-h')
@click.version_option(
    version=__version__,
    prog_name='globus-connect-server-config')
@click.argument('STORAGE_GATEWAY_ID', required=False)
def storage_gateway_show(storage_gateway_id):
    with app.app_context():
        first = True
        query = {}
        try:
            if storage_gateway_id is not None:
                query = {'id': storage_gateway_id}
            for s in StorageGateway.query.filter_by(**query).all():
                if first:
                    first = False
                else:
                    click.echo('')
                if s.connector.display_name == 'POSIX':
                    storage_gateway_show_posix(s)
                elif s.connector.display_name == 'Google Drive':
                    storage_gateway_show_google_drive(s)
                elif s.connector.display_name == 'S3':
                    storage_gateway_show_s3(s)
                elif s.connector.display_name == 'Ceph':
                    storage_gateway_show_ceph(s)
                elif s.connector.display_name == 'BlackPearl':
                    storage_gateway_show_blackpearl(s)
                elif s.connector.display_name == 'Box':
                    storage_gateway_show_box(s)
                elif s.connector.display_name == 'Google Cloud Storage':
                    storage_gateway_show_google_cloud_storage(s)
                else:
                    raise Exception('Unsupported connector {0}'.format(
                        s.connector.display_name))
        except OperationalError:
            click.echo('Permission denied')


if __name__ == '__main__':
    storage_gateway_show()
