from __future__ import absolute_import
import click
import globus_sdk
import json

from .constants import (
    POSIX_CONNECTOR_ID, GDRIVE_CONNECTOR_ID, S3_CONNECTOR_ID,
    CEPH_CONNECTOR_ID, BOX_CONNECTOR_ID, GOOGLE_CLOUD_STORAGE_CONNECTOR_ID,
    CONNECTOR_DISPLAY_NAMES, BLACKPEARL_CONNECTOR_ID
)

from ..app import app, db
from globus.manager.api.utils.services import get_transfer_client
from ..utils import BadInvocation, check_subscription
from globus.manager.api.connectors.box import BoxConnectionError, BoxConnector
from globus.manager.api.models import (
    Collection,
    Configuration,
    StorageGateway,
    UserCredential,
    Connector,
)
from globus.manager.version import __version__
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import OperationalError

posix_options = [
    ('users_allow', click.option(
        '--users-allow', '-u',
        metavar='USERS',
        help='POSIX users allowed to create collections')),
    ('users_deny', click.option(
        '--users-deny', '-U',
        metavar='USERS',
        help='POSIX users not allowed to create collections')),
    ('groups_allow', click.option(
        '--groups-allow', '-g',
        metavar='GROUPS',
        help='POSIX groups allowed to create collections')),
    ('groups_deny', click.option(
        '--groups-deny', '-G',
        metavar='GROUPS',
        help='POSIX groups not allowed to create collections')),
]

client_id_help = 'Google Drive app client id'
client_secret_help = 'Google Drive app client secret'
google_cloud_storage_options = []

if 'Google Cloud Storage' in CONNECTOR_DISPLAY_NAMES:
    client_id_help = 'Google Drive or Google Cloud Storage app client id'
    client_secret_help = 'Google Drive or Google Cloud Storage ' \
                         'app client secret'
    google_cloud_storage_options = [
        ('google_cloud_storage_bucket', click.option(
            '--google-cloud-storage-bucket', '--gb',
            multiple=True, metavar='BUCKET',
            help=('Google Cloud Storage bucket to use. To use multiple '
                  'buckets, pass this option multiple times. Defaults to '
                  'all credential accessible buckets if none specified.'))),
        ('google_cloud_storage_project', click.option(
            '--google-cloud-storage-project', '--gp',
            multiple=True, metavar='PROJECT_ID',
            help=('Google Cloud Storage project to restrict access to. '
                  'To allow multiple projects, pass this option multiple '
                  'times.'))),
    ]

google_options = [
    ('client_id', click.option(
        '--client-id', '-c',
        metavar='ID',
        help=client_id_help)),
    ('client_secret', click.option(
        '--client-secret', '-S',
        metavar='SECRET',
        help=client_secret_help)),
]

google_drive_options = [
    ('user_api_rate_quota', click.option(
        '--user-api-rate-quota', '-q',
        metavar='QUOTA',
        help='Google Drive API rate quota',
        type=click.IntRange(min=1))),
]

s3_ceph_and_blackpearl_options = [
    ('s3_endpoint', click.option(
        '--s3-endpoint',
        metavar='URL',
        help='Full URL of S3 compatible storage location. May include a port.'
             'e.g. https://s3.amazonaws.com:443')),
]

s3_and_ceph_options = [
    ('s3_bucket', click.option(
        '--s3-bucket', '-b', multiple=True, metavar='BUCKET',
        help=('S3 compatible bucket to use. To use multiple buckets, pass '
              'this option multiple times. Defaults to all credential '
              'accessible buckets if none specified.'))),
]

s3_only_options = [
    ('s3_user_credential', click.option(
        '--s3-user-credential/--s3-unauthenticated', default=None,
        help='Either use S3 compatible user credentials for access or use '
             'unauthenticated access. Defaults to user credential access.')),
]

ceph_only_options = [
    ('ceph_admin_key_id', click.option(
        '--ceph-admin-key-id',
        metavar='ID',
        help='Admin key ID for Ceph access')),
    ('ceph_admin_secret_key', click.option(
        '--ceph-admin-secret-key',
        metavar='SECRET',
        help='Admin secret key for Ceph access')),
]

blackpearl_only_options = [
    ('bp_access_id_file', click.option(
        '--bp-access-id-file',
        metavar='FILENAME',
        help='Path to file that maps users to BlackPearl credentials')),
]

box_options = [
    ('box-json-config', click.option(
        '--box-json-config',
        type=click.File('rb'),
        help='The JSON config file downloaded from Box')),
]

common_options = [
    ('display_name', click.option(
        '--display-name', '-n',
        metavar='NAME',
        help='Storage gateway name')),
    ('domain', click.option(
        '--domain', '-d',
        metavar='DOMAIN',
        multiple=True,
        help='Allowed username domain')),
    ('identity_provider', click.option(
        '--identity-provider', '-i',
        metavar='ID',
        multiple=True,
        help='Allowed identity provider')),
    ('restrict_paths', click.option(
        '--restrict-paths', '-p',
        metavar='RP',
        help='Sharing path restrictions')),
    ('allow_guest_collections', click.option(
        '--allow-guest-collections/--disallow-guest-collections',
        default=None,
        help='Dis/allow guest collections on this gateway. '
             'New storage gateways allow by default.')),
    ('allow_mapped_collections', click.option(
        '--allow-mapped-collections/--disallow-mapped-collections',
        default=None,
        help='Dis/allow mapped collections on this gateway. '
             'New storage gateways disallow by default.')),
    ('authentication-assurance-timeout', click.option(
        '--authentication-assurance-timeout',
        type=click.IntRange(min=1),
        metavar='TIMEOUT',
        default=None,
        help='Assure authentication has occurred within the last '
        'TIMEOUT minutes')),
]


def add_options(options):
    def _add_options(func):
        for key, option in options:
            func = option(func)
        return func
    return _add_options


def domain_or_idp_change_is_safe(
        auth_client, sg_id, domains, idps, explicit_idps=False):
    user_credentials = UserCredential.query.filter_by(
        storage_gateway_id=sg_id).all()
    uc_ids = ",".join(list(set([
        (uc.identity_id)
        for uc in user_credentials
        if uc.identity_id != ':globus-mapping:'])))
    bad_domains = {}
    bad_idps = {}
    if uc_ids != "":
        for identity in auth_client.get_identities(
                ids=uc_ids)['identities']:
            username = identity['username']
            domain = username.rsplit('@', 1)[1]
            idp = identity['identity_provider']
            if domains != [] and domain not in domains:
                if domain not in bad_domains:
                    bad_domains[domain] = []
                bad_domains[domain].append(identity)
            if idps != [] and idp not in idps:
                if idp not in bad_idps:
                    bad_idps[idp] = []
                bad_idps[idp].append(identity)
    domain_affected_collections = set()
    for d in bad_domains.keys():
        for bdi in bad_domains[d]:
            affected_user_credentials = [
                (uc.id) for uc in user_credentials
                if uc.identity_id == bdi['id']
            ]
            domain_affected_collections.update(
                set([c.id for c in Collection.query.filter(
                    Collection.user_credential_id.in_(
                        affected_user_credentials)).all()]))
    idp_affected_collections = set()
    for i in bad_idps.keys():
        for bii in bad_idps[i]:
            affected_user_credentials = [
                (uc.id) for uc in user_credentials
                if uc.identity_id == bii['id']]
            idp_affected_collections.update(
                set([c.id for c in Collection.query.filter(
                    Collection.user_credential_id.in_(
                        affected_user_credentials)).all()]))
    affected_collections = set()

    if len(domain_affected_collections) > 0:
        affected_collections.update(domain_affected_collections)
    if explicit_idps and len(idp_affected_collections) > 0:
        affected_collections.update(idp_affected_collections)
    if len(affected_collections) > 0:
        raise BadInvocation('''\
Removing the ID providers or domains on a storage gateway is only supported
if there are no existing collections which were created with policies
depending on those ID providers or domains.

Some existing collections prevent this change. Please delete these
collections with gcs-config collection delete ID...
or pass multiple --domain and/or --identity provider options to this script
to include both the old and new configurations.
(eg. gcs-config storage-system update -d domain1 -d domain2)

The affected collections are
- {}'''.format(
            "\n- ".join(affected_collections)))


def gateway_subscription_check(**kwargs):
    """
    Determines if the given set of options requires a subscription.
    If so, calls check_subscription with a reason for why
    a subscription is needed if the check fails.

    Must be run in an app context
    """
    require_subscription = False

    if kwargs.get('allow_guest_collections'):
        require_subscription = (
            'to configure a storage gateway for guest '
            'collections')

    if kwargs.get('connector'):
        connector = Connector.query.filter_by(
            display_name=kwargs['connector']).one()
        if connector.id != POSIX_CONNECTOR_ID:
            require_subscription = 'to use {} connector'.format(
                connector.display_name)

    if (kwargs.get('high_assurance')
            or kwargs.get('authentication_assurance_timeout')):
        require_subscription = (
            'to create a high-assurance storage gateway')

    if require_subscription is not False:
        host_endpoint_id = Configuration.query.filter_by(
            name='host_endpoint_id').one().value
        check_subscription(host_endpoint_id, reason=require_subscription)


def storage_gateway_update_storage(sg, **kwargs):
    storage_policies = sg.storage_policies
    dsi_config = storage_policies['private'].get('dsi_config') or {}
    domains = kwargs.get('domain')
    idp = kwargs.get('identity_provider')
    transfer_client = get_transfer_client()
    client_id = Configuration.query.filter_by(
        name='client_id').one().value
    client_secret = Configuration.query.filter_by(
        name='client_secret').one().value
    auth_client = globus_sdk.ConfidentialAppAuthClient(
        client_id, client_secret)
    sg_user_credentials = UserCredential.query.join(
        StorageGateway).filter(StorageGateway.id == sg.id).all()
    sg_collections = Collection.query.join(UserCredential).join(
        StorageGateway).filter(StorageGateway.id == sg.id).all()

    if domains != ():
        if domains == ('',):
            # Removing domain restrictions will not invalidate any
            # user creds and collections
            if 'domains' in storage_policies['public']:
                del storage_policies['public']['domains']
        else:
            if idp == ():
                idps = auth_client.get(
                    '/v2/api/identity_providers?domains={}'.format(
                        ','.join(domains)))['identity_providers']

                if not idps:
                    raise BadInvocation(
                        'Unable to resolve any ID providers for domain(s) {}'
                        .format(','.join(domains)))

                idp_ids = [i['id'] for i in idps]
                domain_or_idp_change_is_safe(
                    auth_client,
                    sg.id,
                    list(domains),
                    idp_ids,
                    explicit_idps=False)
                storage_policies['public']['identity_providers'] = [
                    i['id'] for i in idps]
            else:
                domain_or_idp_change_is_safe(
                    auth_client,
                    sg.id,
                    list(domains),
                    list(idp),
                    explicit_idps=True)
            storage_policies['public']['domains'] = domains

    # confirm any domain changes would not violate high assurance reqs
    if (sg.require_high_assurance
            and len(storage_policies['public'].get('domains', [])) != 1):
        click.echo(
            'High assurance gateways must '
            'be configured with a single required user domain')
        click.get_current_context().exit(1)

    if idp != ():
        if idp == ('',):
            if 'identity_providers' in storage_policies['public']:
                del storage_policies['public']['identity_providers']
        else:
            domain_or_idp_change_is_safe(
                auth_client, sg.id, [], list(idp), explicit_idps=True)
            storage_policies['public']['identity_providers'] = list(idp)

    if kwargs.get('restrict_paths') is not None:
        storage_policies['private']['restrict_paths'] = (
            kwargs['restrict_paths'])

    # confirm no mismatched connector options were used
    for connector_ids, option_set in [
        ((POSIX_CONNECTOR_ID), posix_options),
        ((GDRIVE_CONNECTOR_ID,
          GOOGLE_CLOUD_STORAGE_CONNECTOR_ID), google_options),
        ((GDRIVE_CONNECTOR_ID), google_drive_options),
        ((GOOGLE_CLOUD_STORAGE_CONNECTOR_ID), google_cloud_storage_options),
        ((S3_CONNECTOR_ID), s3_only_options),
        ((CEPH_CONNECTOR_ID), ceph_only_options),
        ((S3_CONNECTOR_ID, CEPH_CONNECTOR_ID,
          BLACKPEARL_CONNECTOR_ID), s3_ceph_and_blackpearl_options),
        ((S3_CONNECTOR_ID, CEPH_CONNECTOR_ID), s3_and_ceph_options),
        ((BLACKPEARL_CONNECTOR_ID), blackpearl_only_options),
        ((BOX_CONNECTOR_ID), box_options),
    ]:
        if sg.connector.id not in connector_ids:
            for (option, _) in option_set:
                if kwargs.get(option) is not None and kwargs.get(option) != ():
                    click.echo(
                        'Error: storage option "{}" used with incorrect '
                        'connector {} '.format(
                            option, sg.connector.display_name))
                    click.get_current_context().exit(1)

    # Google credential
    if sg.connector.id in (GDRIVE_CONNECTOR_ID,
                           GOOGLE_CLOUD_STORAGE_CONNECTOR_ID):

        if sg.credential:
            existing_client_id = sg.credential.get('client_id')
            existing_client_secret = sg.credential.get('secret')
        else:
            existing_client_id = None
            existing_client_secret = None

        client_id = kwargs.get('client_id') or existing_client_id
        if client_id is None:
            raise BadInvocation(
                'Storage gateways using the {} connector require a --client-id'
                .format(sg.connector.display_name))

        secret = kwargs.get('client_secret') or existing_client_secret
        if secret is None:
            raise BadInvocation(
                'Storage gateways using the {} connector require a '
                '--client-secret'.format(sg.connector.display_name))

        sg.credential = {
            'secret': secret,
            'client_id': client_id,
        }

    # Google Drive specific policies
    if sg.connector.id == GDRIVE_CONNECTOR_ID:
        load_dsi_module = 'google_drive'
        if kwargs.get('user_api_rate_quota') is not None:
            load_dsi_module += ":api_quota_rate={0}".format(
                kwargs['user_api_rate_quota'])
        storage_policies['private']['process_user'] = 'gcsweb'
        storage_policies['private']['load_dsi_module'] = load_dsi_module

    # POSIX specific policies
    if sg.connector.id == POSIX_CONNECTOR_ID:
        access_types = [
            'users_allow', 'users_deny', 'groups_allow', 'groups_deny']
        for access in access_types:
            val = kwargs.get(access)
            if val is not None:
                if val != '':
                    storage_policies['private'][access] = val
                else:
                    del storage_policies['private'][access]

    # S3 and Ceph policies
    if sg.connector.id in (S3_CONNECTOR_ID,
                           CEPH_CONNECTOR_ID,
                           BLACKPEARL_CONNECTOR_ID):

        s3_endpoint = (kwargs.get('s3_endpoint') or
                       dsi_config.get('s3_endpoint'))

        if not s3_endpoint:
            click.echo('Storage gateways using the {} connector require '
                       '--s3-endpoint'.format(sg.connector.display_name))
            click.get_current_context().exit(1)

        if not (s3_endpoint.startswith("http://") or
                s3_endpoint.startswith("https://")):
            click.echo('--s3-endpoint must begin with http:// or https://')
            click.get_current_context().exit(1)

        if sg_collections:
            click.echo('Warning! changing the S3 endpoint on this gateway may '
                       'break existing collections. Previous S3 endpoint was: '
                       '{}'.format(dsi_config['s3_endpoint']))
        dsi_config['s3_endpoint'] = s3_endpoint

        storage_policies['private']['process_user'] = 'gcsweb'

    # S3 and Ceph policies
    if sg.connector.id in (S3_CONNECTOR_ID, CEPH_CONNECTOR_ID):

        s3_buckets = kwargs.get('s3_bucket')
        if s3_buckets != ():
            dsi_config['s3_buckets'] = s3_buckets or []
        if dsi_config.get('s3_buckets') is None:
            dsi_config['s3_buckets'] = []

    # S3 specific policies
    if sg.connector.id == S3_CONNECTOR_ID:
        storage_policies['private']['load_dsi_module'] = 's3'

        s3_user_credential = kwargs.get('s3_user_credential')
        if s3_user_credential is not None:
            if sg_user_credentials:
                click.echo(
                    'Cannot update S3 access method on this gateway as it '
                    'has existing user credentials.')
                click.get_current_context().exit(1)

            storage_policies['public'][
                's3_user_credential_required'] = s3_user_credential
            dsi_config['s3_user_credential_available'] = s3_user_credential

        # default to requiring user credentials on creation
        if storage_policies['public'].get(
                's3_user_credential_required') is None:
            storage_policies['public']['s3_user_credential_required'] = True
            dsi_config['s3_user_credential_available'] = True

    # Ceph specific policies
    if sg.connector.id == CEPH_CONNECTOR_ID:
        storage_policies['private']['load_dsi_module'] = 'ceph'

        ceph_admin_key_id = (kwargs.get('ceph_admin_key_id') or
                             dsi_config.get('ceph_admin_key_id'))
        if not ceph_admin_key_id:
            click.echo('Storage gateways using the Ceph connector require a '
                       '--ceph-admin-key-id')
            click.get_current_context().exit(1)
        dsi_config['ceph_admin_key_id'] = ceph_admin_key_id

        ceph_admin_secret_key = (kwargs.get('ceph_admin_secret_key') or
                                 dsi_config.get('ceph_admin_secret_key'))
        if not ceph_admin_secret_key:
            click.echo('Storage gateways using the Ceph connector require a '
                       '--ceph-admin-secret-key')
            click.get_current_context().exit(1)
        dsi_config['ceph_admin_secret_key'] = ceph_admin_secret_key

    # BlackPearl specific policies
    if sg.connector.id == BLACKPEARL_CONNECTOR_ID:
        storage_policies['private']['load_dsi_module'] = 'blackpearl'

        access_id_file = (kwargs.get('bp_access_id_file') or
                          dsi_config.get('bp_access_id_file'))
        if not access_id_file:
            click.echo('Storage gateways using the BlackPearl '
                       'connector require a --bp-access-id-file')
            click.get_current_context().exit(1)
        dsi_config['bp_access_id_file'] = access_id_file

    # Box specific policies
    if sg.connector.id == BOX_CONNECTOR_ID:
        storage_policies['private']['load_dsi_module'] = 'box'
        storage_policies['private']['process_user'] = 'gcsweb'

        # get the path for the downloaded JSON config file
        box_json_config = kwargs.get('box_json_config')
        existing_config = storage_policies['private'].get('dsi_config', {})

        if not (box_json_config or existing_config):
            click.echo('Storage gateways using the Box connector require '
                       '--box-json-config')
            click.get_current_context().exit(1)

        if box_json_config:
            # try to parse the JSON file
            try:
                dsi_config = json.load(box_json_config)
                storage_policies['private']['dsi_config'] = dsi_config
            except ValueError:
                click.echo("Unable to parse JSON from Box JSON "
                           "config file {}".format(box_json_config))
                click.get_current_context().exit(1)

        # confirm we can get an enterprise access token with the dsi config
        try:
            BoxConnector.get_new_enterprise_access_token(storage_policies)
        except BoxConnectionError:
            click.echo("Unable to connect to Box with JSON config")
            click.get_current_context().exit(1)

    # Google Cloud Storage specific policies
    if sg.connector.id == GOOGLE_CLOUD_STORAGE_CONNECTOR_ID:

        storage_policies['private']['load_dsi_module'] = 'google_cloud'
        storage_policies['private']['process_user'] = 'gcsweb'

        # google_cloud_storage_buckets
        gb = kwargs.get('gb')
        if gb != ():
            dsi_config[
                'google_cloud_storage_buckets'] = gb or []
        if dsi_config.get('google_cloud_storage_buckets') is None:
            dsi_config['google_cloud_storage_buckets'] = []

        # google_cloud_storage_projects
        gp = kwargs.get('gp')
        if gp != ():
            dsi_config[
                'google_cloud_storage_projects'] = gp or []
        if dsi_config.get('google_cloud_storage_projects') is None:
            dsi_config['google_cloud_storage_projects'] = []

        # confirm that changing google_cloud_storage_projects does not make
        # existing collections unusable
        if dsi_config['google_cloud_storage_projects']:
            affected_collections = []
            for col in sg_collections:
                if col.collection_policies.get('private').get(
                    'google_cloud_storage_project') not in dsi_config[
                        'google_cloud_storage_projects']:
                    affected_collections.append(col.id)

            if affected_collections:
                raise BadInvocation('''\
Removing projects from a Goggle Cloud Storage gateway is only supported
if there are no existing collections which were created with policies
depending on those projects.

Some existing collections prevent this change. Please delete these
collections with gcs-config collection delete ID...
or pass multiple --google-cloud-storage-projects options to this script.
(eg. gcs-config storage-system update --gp project-1 --gp project-2)

The affected collections are
- {}'''.format("\n- ".join(affected_collections)))

    # set display name
    display_name = kwargs.get('display_name')
    if display_name is not None:
        sg.display_name = display_name

    # set allow_mapped_collections and allow_guest_collections
    if kwargs.get("allow_guest_collections") is not None:
        allow_guest_collections = kwargs.get("allow_guest_collections")
    else:
        if sg.allow_guest_collections is not None:
            allow_guest_collections = sg.allow_guest_collections
        else:
            allow_guest_collections = True  # on by default

    if kwargs.get("allow_mapped_collections") is not None:
        allow_mapped_collections = kwargs.get("allow_mapped_collections")
    else:
        if sg.allow_mapped_collections is not None:
            allow_mapped_collections = sg.allow_mapped_collections
        else:
            allow_mapped_collections = False  # off by default

    # storage gateway must have at least one collection type allowed
    if not (allow_guest_collections or allow_mapped_collections):
        click.echo("Storage gateways must allow at "
                   "least one collection type.")
        click.get_current_context().exit(1)

    # if mapped collections are being enabled, confirm the connector
    # also allows mapped collections
    if (allow_mapped_collections and
            (not sg.connector.allow_mapped_collections)):
        click.echo("Connector {} cannot support mapped collections."
                   .format(sg.connector.display_name))
        click.get_current_context().exit(1)

    # if a type is being disabled, confirm no such collections already exist
    if (kwargs.get("allow_guest_collections") is False
            or kwargs.get("allow_mapped_collections") is False):

        for collection in sg_collections:
            if (not allow_guest_collections and
                    collection.collection_policies["public"]["type"]
                    == "guest"):
                click.echo("Unable to disable guest collections on a storage "
                           "gateway with existing guest collections.")
                click.get_current_context().exit(1)

            if (not allow_mapped_collections and
                    collection.collection_policies["public"]["type"]
                    == "mapped"):
                click.echo("Unable to disable mapped collections on a storage "
                           "gateway with existing mapped collections.")
                click.get_current_context().exit(1)

    # set or update require_high_assurance and the
    # authentication_assurance_timeout
    if kwargs.get('authentication_assurance_timeout'):
        # check that the domain policies are correct for high assurance
        if len(storage_policies['public'].get('domains', [])) != 1:
            click.echo(
                'High assurance gateways must '
                'be configured with a single required user domain')
            click.get_current_context().exit(1)

        # update any existing collection's session timeouts in transfer
        for collection in sg_collections:
            update_data = {
                "authentication_assurance_timeout": kwargs[
                    "authentication_assurance_timeout"] * 60
            }
            transfer_client.update_endpoint(collection.id, update_data)

        # update public policies
        storage_policies['public']['authentication_assurance_timeout'] = (
            kwargs['authentication_assurance_timeout'])

        # set high assurance to True, this must never be unset
        sg.require_high_assurance = True

    if allow_mapped_collections and not sg.require_high_assurance:
        click.echo("Non-High Assurance mapped collections not supported.")
        click.get_current_context().exit(1)

    sg.allow_mapped_collections = allow_mapped_collections
    sg.allow_guest_collections = allow_guest_collections
    storage_policies['private']['dsi_config'] = dsi_config
    sg.storage_policies = storage_policies

    db.session.add(sg)
    db.session.commit()


SHARED_OPTION_HELP_TEXT = '''
Additional parameters control the policies for the storage gateway:
{0}

Each available connector type has several additional options to control
connector-specific policies:

POSIX:
{1}

Google Drive:
{2}

S3:
{3}

Ceph:
{4}

Box:
{5}

BlackPearl:
{6}
'''.format(
    ', '.join([
        '--{}'.format(x[0].replace('_', '-'))
        for x in common_options
        if x[0] not in ['display_name']
    ]),
    ', '.join([
        '--{}'.format(x[0].replace('_', '-'))
        for x in posix_options
    ]),
    ', '.join([
        '--{}'.format(x[0].replace('_', '-'))
        for x in google_drive_options
    ]),
    ', '.join([
        '--{}'.format(x[0].replace('_', '-'))
        for x in (s3_only_options
                  + s3_ceph_and_blackpearl_options
                  + s3_and_ceph_options)
    ]),
    ', '.join([
        '--{}'.format(x[0].replace('_', '-'))
        for x in (ceph_only_options
                  + s3_ceph_and_blackpearl_options
                  + s3_and_ceph_options)
    ]),
    ', '.join([
        '--{}'.format(x[0].replace('_', '-'))
        for x in box_options
    ]),
    ', '.join([
        '--{}'.format(x[0].replace('_', '-'))
        for x in blackpearl_only_options + s3_ceph_and_blackpearl_options
    ]),
)


@click.command(
    'update',
    short_help='Update a storage gateway',
    help='''\
The storage-gateway update subcommand modifies a storage gateway, which
is a named location for creating collections. It also can associate optional
policies governing path restrictions and identity providers which may access
the storage gateway.

This command does a replace of the storage gateway parameters that are present
on the command-line. Options described below which are not present will retain
their value. To delete a parameter's value, pass the empty string ('') as its
value (e.g. --groups-allow '').

{}'''.format(SHARED_OPTION_HELP_TEXT))
@click.help_option('--help', '-h')
@click.version_option(
    version=__version__,
    prog_name='globus-connect-server-config')
@add_options(common_options)
@add_options(posix_options)
@add_options(google_options)
@add_options(google_drive_options)
@add_options(s3_ceph_and_blackpearl_options)
@add_options(s3_and_ceph_options)
@add_options(s3_only_options)
@add_options(ceph_only_options)
@add_options(blackpearl_only_options)
@add_options(box_options)
@add_options(google_cloud_storage_options)
@click.argument('STORAGE_GATEWAY_ID', required=True)
@click.pass_context
def storage_gateway_update(ctx, **kwargs):
    with app.app_context():
        sgid = kwargs.get('storage_gateway_id')

        try:
            sg = StorageGateway.query.filter_by(id=sgid).one()
            gateway_subscription_check(**kwargs)
            storage_gateway_update_storage(sg, **kwargs)
        except OperationalError:
            click.echo('Permission denied')
            click.get_current_context().exit(1)
        except NoResultFound:
            click.echo('No such storage gateway {}'.format(sgid))
            click.get_current_context().exit(1)
        except BadInvocation as e:
            click.echo(str(e))
            click.get_current_context().exit(1)


if __name__ == '__main__':
    storage_gateway_update()
