from flask import current_app
from globus.manager.api.models import (
    db,
    Configuration,
    PreallocatedCollection
)
from globus.manager.api.utils import virtual_client, https_endpoint
from globus.manager.api.utils.services import get_dns_client
import globus_sdk

import os
import queue
import socket
import subprocess
import sys
import threading
import time

LOW_WATER_MARK = 4
HIGH_WATER_MARK = 10
CHECKS_PER_MINUTE = 4

last_config_check = time.time()
last_config_lock = threading.Lock()
globus_environment = None
gcs_certificate = None
server_name = None


def init_globals():
    def init_global(key, initial):
        val = initial
        with current_app.app_context():
            if val is None:
                val = current_app.config.get(key.upper())
                if val is None:
                    try:
                        val = Configuration.query.filter_by(
                            name=key).one().value
                        current_app.config[key.upper()] = val
                    except Exception as e:
                        current_app.logger.error(
                            "Unable to retrieve {0} from db: {1}".format(
                                key, e))
        return val

    _globus_environment = init_global('environment', globus_environment)
    _gcs_certificate = init_global('gcs_certificate', gcs_certificate)
    _server_name = init_global('server_name', server_name)

    if _globus_environment is None or _gcs_certificate is None:
        with last_config_lock:
            global last_config_check
            since_last = time.time() - last_config_check
            if since_last < 30.0:
                print("Sleeping {0}".format(30.0 - since_last))
                time.sleep(30.0 - since_last)
            last_config_check = time.time()
            return None
    return (_server_name, _globus_environment, _gcs_certificate)


class PreallocatedCollectionState(object):
    def __init__(self, server_name=None):
        self.auth_id = None
        self.scopes = None
        self.dns_name = None
        self.certificate_path = None
        self.https_config = None
        self.server_name = server_name
        self.collection = PreallocatedCollection()

    def dns_name_create(self):
        current_app.logger.debug("enter: dns_name_create")
        with current_app.app_context():
            try:
                dns_client = get_dns_client()
            except globus_sdk.exc.AuthAPIError as e:
                if e.http_status == 401:
                    print(
                        'FATAL: Unable to authorize DNS client: '
                        '{}'.format(e))
                    sys.exit(1)
                else:
                    print('WARNING: Exception: {}'.format(e))
                    time.sleep(10)
                    raise(e)

        domain_response = dns_client.create_domain(cname=self.server_name)
        self.dns_name = domain_response.data['domain']

        resolved = False
        attempt = 0
        while not resolved and attempt < 10:
            try:
                socket.gethostbyname(self.dns_name)
                resolved = True
            except socket.gaierror:
                time.sleep(30)
            attempt = attempt + 1

        current_app.logger.debug("exit: dns_name_create")
        return self

    def virtual_client_register(self, migration_uuid=None):
        current_app.logger.debug("enter: virtual_client_register")
        self.collection.domain = self.dns_name
        client, scopes, fqdn = virtual_client.register(
            self.collection, migration_uuid=migration_uuid)
        self.auth_id = client['id']
        self.scopes = scopes
        self.dns_name = fqdn
        current_app.logger.debug("exit: virtual_client_register")
        return self

    def certificate_get(self):
        current_app.logger.debug("enter: certificate_get")
        certbot_args = [
            '/opt/globus/bin/certbot_wrapper',
            '--quiet',
            '--domain', self.dns_name,
        ]
        certbot_wrapper = subprocess.Popen(certbot_args)
        certbot_wrapper.communicate()
        ret = certbot_wrapper.returncode
        if ret != 0:
            raise Exception('Failed to create letsencrypt certificate')
        self.certificate_path = os.path.join(
            os.path.expanduser('~/etc/letsencrypt'),
            'live',
            self.dns_name,
            'cert.pem')

        current_app.logger.debug("exit: certificate_get")
        return self

    def https_config_write(self):
        current_app.logger.debug("enter: https_config_write")
        try:
            self.https_endpoint = https_endpoint.HTTPSEndpoint(
                self.auth_id,
                current_app.config['CLIENT_ID'],
                current_app.config['CLIENT_SECRET'],
                self.dns_name,
                environment=os.getenv('GLOBUS_SDK_ENVIRONMENT', 'production'))
            self.https_endpoint.setup()
        except Exception as e:
            current_app.logger.exception(e)
            sys.exit(1)
        current_app.logger.debug("exit: https_config_write")
        return self

    def collection_commit(self):
        current_app.logger.debug("enter: collection_commit")
        with current_app.app_context():
            db.session.add(self.collection)
            db.session.commit()
        current_app.logger.debug("exit: collection_commit")
        return self


class PipelineStage(threading.Thread):
    def __init__(self, in_queue, out_queue, work, *kwargs):
        self.in_queue = in_queue
        self.out_queue = out_queue
        self.work = work
        super(PipelineStage, self).__init__(*kwargs)

    def run(self):
        while True:
            item = self.in_queue.get()
            try:
                next_item = self.work(item)
                if next_item is not None:
                    self.out_queue.put(next_item)
                else:
                    self.in_queue.put(item)
            except Exception:
                self.in_queue.put(item)
            finally:
                self.in_queue.task_done()


class Pipeline(object):
    def __init__(self, queue_type=queue.Queue, maxsize=0, stages=[]):
        self.pipeline = []
        self.maxsize = maxsize
        self.work_queue = queue_type(maxsize=maxsize)

        for stage in stages:
            self.add_stage(stage)

    def add_stage(self, work):
        if len(self.pipeline) == 0:
            in_queue = queue.Queue()
        else:
            in_queue = self.pipeline[-1].out_queue
        out_queue = queue.Queue()
        self.pipeline.append(PipelineStage(in_queue, out_queue, work))
        self.pipeline[-1].start()

    def put(self, work=None):
        self.work_queue.put(work)
        self.pipeline[0].in_queue.put(self.work_queue.get())

    def get(self):
        final = self.pipeline[-1].out_queue.get()
        self.pipeline[-1].out_queue.task_done()
        self.work_queue.task_done()
        return final


class CollectionPreallocator(object):
    def run(self):
        global globus_environment, gcs_certificate, server_name

        while (globus_environment is None
               or gcs_certificate is None
               or server_name is None):
            res = init_globals()
            if res is not None:
                server_name, globus_environment, gcs_certificate = res

        if globus_environment.lower() != 'production':
            global LOW_WATER_MARK, HIGH_WATER_MARK
            LOW_WATER_MARK = 1
            HIGH_WATER_MARK = 2

        os.environ["GLOBUS_SDK_ENVIRONMENT"] = globus_environment.lower()

        log_level = os.environ.get('GCS_MANAGER_ASSISTANT_LOG_LEVEL')
        if log_level:
            current_app.logger.setLevel(log_level)

        with current_app.app_context():
            pipeline = Pipeline(
                maxsize=HIGH_WATER_MARK,
                stages=[
                    PreallocatedCollectionState.dns_name_create,
                    PreallocatedCollectionState.virtual_client_register,
                    PreallocatedCollectionState.certificate_get,
                    PreallocatedCollectionState.https_config_write,
                    PreallocatedCollectionState.collection_commit,
                ])
            in_process = 0
            while True:
                current = PreallocatedCollection.query.count()
                while (current < LOW_WATER_MARK
                        and (current + in_process) < HIGH_WATER_MARK
                        and in_process < pipeline.maxsize):
                    pipeline.put(PreallocatedCollectionState(
                        server_name=server_name))
                    in_process = in_process + 1
                if in_process == 0:
                    time.sleep(60.0 / CHECKS_PER_MINUTE)
                else:
                    if in_process > 0:
                        current_app.logger.info(
                            "Processing {} collections".format(
                                in_process))
                    while in_process > 0:
                        pipeline.get()
                        in_process = in_process - 1
                        current = PreallocatedCollection.query.count()
