#! /opt/globus/bin/python
"""
Globus Connect Server cleanup
"""

from __future__ import absolute_import

import logging
import os
import socket
import stat

import click
import globus_sdk

from globus.manager.version import __version__

from globus.manager.api.utils.services import (
    DNS_SCOPE,
    get_dns_client,
)
from globus.manager.scripts.gcs_config.collection.delete import (
    app,
    collection_delete,
    db,
)

from globus.manager.api.models import (
    Configuration,
    PreallocatedCollection,
    StorageGateway,
    UserCredential,
)

from .. import (
    ConfigFile,
    EndpointManager,
    GCSManager,
    GridFTPService,
    ServiceManager,
    WebServerManager,
    log as logger,
)
from globus.manager.utils.platform_info import (
    PlatformInfo,
    UnsupportedPlatformException
)

RETRIES = 15
dns_name_file = '/var/lib/globus-connect-server/dns-name.txt'
endpoint_uuid_file = '/var/lib/globus-connect-server/endpoint-uuid.txt'


def lookup_domain():
    domain = None
    if os.path.exists(dns_name_file):
        logger.debug(
            "Checking for existing domain in file {0}".format(
                dns_name_file))
        with open(dns_name_file, 'r') as f:
            domain = f.read().strip()
            if domain == '':
                domain = None
            else:
                logger.debug("Using domain {0}".format(domain))
    return domain


def unregister_domain(auth_client, environment='Production'):
    # Delete the registered domain
    domain = lookup_domain()

    if domain is not None:
        dns_authorizer = globus_sdk.ClientCredentialsAuthorizer(
            auth_client, DNS_SCOPE)
        dns_client = get_dns_client(
            authorizer=dns_authorizer, environment=environment.lower())
        try:
            dns_client.delete_domain(domain=domain)
        except Exception as e:
            logger.warning(str(e))

    if os.path.exists(dns_name_file):
        os.remove(dns_name_file)


def delete_virtual_clients(auth_client, parent_client_id):
    '''
    Delete all virtual clients that are owned by this client_id
    '''
    clients = []
    try:
        result = auth_client.get('/v2/api/clients')
        clients = result.data.get('clients')
    except globus_sdk.GlobusAPIError as e:
        logger.warn(e)

    for client in clients:
        try:
            client_id = client['id']
            if client['id'] != parent_client_id:
                logger.info("Deleting virtual client {}".format(client_id))
                auth_client.delete('/v2/api/clients/{}'.format(client_id))
        except Exception as e:
            logger.warn(e)


def remove_letsencrypt(domain):
    letsencryptdir = os.path.expanduser('~gcsweb/etc/letsencrypt')
    livedir = os.path.join(letsencryptdir, 'live', domain)
    archivedir = os.path.join(letsencryptdir, 'archive', domain)
    renewalfile = os.path.join(
        letsencryptdir,
        'renewal',
        '{}.conf'.format(domain))
    if os.path.exists(livedir):
        for p in os.listdir(livedir):
            full = os.path.join(livedir, p)
            if os.path.exists(full):
                os.remove(full)
        os.rmdir(livedir)
    if os.path.exists(archivedir):
        for p in os.listdir(archivedir):
            full = os.path.join(archivedir, p)
            if os.path.exists(full):
                os.remove(full)
        os.rmdir(archivedir)
    if os.path.exists(renewalfile):
        os.remove(renewalfile)


@click.command(
    'globus-connect-server-cleanup',
    short_help='Set up Globus Connect Server services',
    help='''\
The globus-connect-server-cleanup command performs the initial cleanup
of the GCS-related services. It creates a GCS endpoint, and then
enables the services needed to interact with the endpoint.

Before running this command, update the /etc/globus-connect-server.conf
file to, at a minimum, contain values for

- [Globus].ClientId

- [Globus].ClientSecret

- [Endpoint].Name

- [LetsEncrypt].Email

- [LetsEncrypt].AgreeToS''')
@click.option(
    '--config-file', '-c',
    help='Use CONFIG_FILE instead of /etc/globus-connect-server.conf',
    metavar='CONFIG_FILE',
    default='/etc/globus-connect-server.conf',
    type=click.Path(readable=True))
@click.version_option(
    version=__version__,
    prog_name='globus-connect-server-cleanup')
@click.option(
    '--verbose', '-v',
    help='Print more information about tasks',
    is_flag=True)
@click.option(
    '--root', '-r',
    help='Add ROOT as the directory prefix for the configuration',
    metavar='ROOT',
    default='/',
    type=click.Path(dir_okay=True, file_okay=False))
@click.option(
    '--force', '-f',
    help='Ignore errors and existing configuration',
    is_flag=True)
@click.option(
    '--delete', '-d',
    help='Delete the endpoint',
    is_flag=True)
@click.option(
    '--log', '-l',
    help='Log execution information to LOG',
    metavar='LOG',
    default='/var/log/globus-connect-server.log',
    type=click.Path(writable=True, exists=False))
@click.option(
    '--endpoint-uuid', '-u',
    help='Modify the endpoint with the specified UUID',
    metavar='UUID')
@click.pass_context
def main(ctx, config_file, verbose, root, force, delete, log, endpoint_uuid):
    handler = logging.StreamHandler()
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        handler.setLevel(logging.WARNING)
    logger.addHandler(handler)

    log_path = log
    if root != '/':
        log_path = os.path.join(root, log)

    try:
        old_umask = os.umask(stat.S_IRWXG | stat.S_IRWXO)
        log_file = logging.FileHandler(log_path)
        log_file.setLevel(logging.DEBUG)
        log_file.setFormatter(
            logging.Formatter(
                "%(asctime)s: %(levelname)s: %(message)s",
                datefmt='%Y-%m-%dT%H:%M:%S'))
        logger.addHandler(log_file)
        os.umask(old_umask)
    except Exception:
        pass

    supported_platforms = ['redhat', 'debian']
    if PlatformInfo().platform not in supported_platforms:
        click.echo(str(UnsupportedPlatformException()))
        ctx.exit(1)
    try:
        conf = ConfigFile(config_file=config_file, root=root)
    except Exception as e:
        click.echo('Error parsing configuration: {}'.format(str(e)))
        ctx.exit(1)
    os.environ['GLOBUS_SDK_ENVIRONMENT'] = conf.Globus.Instance().lower()
    os.umask(stat.S_IWGRP | stat.S_IWOTH)
    socket.setdefaulttimeout(300)

    client_id = conf.Globus.ClientId()
    secret = conf.Globus.ClientSecret()

    required_options = {
        '[Globus].ClientId': conf.Globus.ClientId(),
        '[Globus].ClientSecret': conf.Globus.ClientSecret(),
        '[LetsEncrypt].Email': conf.LetsEncrypt.Email(),
        '[LetsEncrypt].AgreeToS': conf.LetsEncrypt.AgreeToS(),
    }
    for o in required_options.keys():
        if not required_options[o]:
            click.echo('Invalid value for configuration option {}'.format(o))
            ctx.exit(1)

    logging.info('Disabling services')
    domain = lookup_domain()
    if domain:
        gridftp = GridFTPService.create_using_conf(
            conf, domain=domain, root=root)
        gridftp.cleanup()
        remove_letsencrypt(domain)
    ServiceManager.disable('gcs_manager_assistant')
    ServiceManager.stop('gcs_manager_assistant')
    ServiceManager.disable('gcs_certbot-renew', periodic=True)
    ServiceManager.stop('gcs_certbot-renew', periodic=True)
    for module in WebServerManager.required_modules:
        WebServerManager.module_enable(module)
    for apacheconf in WebServerManager.required_confs:
        WebServerManager.conf_enable(apacheconf)
    WebServerManager.site_disable(GCSManager.service_name)
    WebServerManager.site_remove(GCSManager.service_name)
    ServiceManager.restart(WebServerManager.service)

    # Delete collections
    logger.info('Deleting collections')
    ctx.invoke(collection_delete, collection_id='all')

    # Delete collections owned by client_id but not in db
    # XXX

    with app.app_context():
        if endpoint_uuid is None:
            host_endpoint_id = Configuration.query.filter_by(
                name='host_endpoint_id').one_or_none()
            if host_endpoint_id is not None:
                endpoint_uuid = host_endpoint_id.value
            else:
                endpoint_uuid = conf.Globus.ClientId()

        logger.info('Removing user credentials')
        for uc in UserCredential.query.all():
            logger.debug('Deleting user_credential {}'.format(uc.display_name))
            db.session.delete(uc)
        logger.info('Removing storage gateways')
        for sg in StorageGateway.query.all():
            logger.debug('Deleting storage_gateway {}'.format(sg.display_name))
            db.session.delete(sg)
        httpdir = os.path.expanduser('~gcsweb/etc/httpd/conf.d')
        for pc in PreallocatedCollection.query.all():
            logger.debug('Deleting preallocated collection {}'.format(pc.id))
            http_conf = os.path.join(httpdir, '{}.conf'.format(pc.id))
            if os.path.exists(http_conf):
                os.remove(http_conf)
            remove_letsencrypt(pc.domain)
            db.session.delete(pc)
        db.session.commit()

    try:
        TRANSFER_SCOPE = 'urn:globus:auth:scope:transfer.api.globus.org:all'
        auth_client = globus_sdk.ConfidentialAppAuthClient(client_id, secret)
        transfer_authorizer = globus_sdk.ClientCredentialsAuthorizer(
            auth_client, TRANSFER_SCOPE)
        api = globus_sdk.TransferClient(
            authorizer=transfer_authorizer)

        if endpoint_uuid is not None:
            logger.info('Cleaning Up Globus Connect Server endpoint')
            endpoint_manager = EndpointManager.create_using_conf(
                conf=conf,
                api=api, domain=domain,
                endpoint_uuid=endpoint_uuid)
            endpoint_manager.cleanup(delete_endpoint=delete)
    except Exception as e:
        logger.warning(str(e))

    # Delete DNS
    logger.info('Deleting DNS entry')
    unregister_domain(auth_client, environment=conf.Globus.Instance())
    delete_virtual_clients(auth_client, client_id)

    # Delete other configuration files
    logger.info('Finishing cleanup')
    if os.path.exists(endpoint_uuid_file):
        os.remove(endpoint_uuid_file)

    with app.app_context():
        logger.info('Removing configuration')
        for c in Configuration.query.all():
            if c.name != 'secret':
                logger.debug('Deleting configuration {}'.format(c.name))
                db.session.delete(c)
        db.session.commit()
