#! /opt/globus/bin/python
"""
Globus Connect Server setup
"""

from __future__ import absolute_import

import glob
import logging
import os
import grp
import pwd
import re
import socket
import stat
import subprocess
import sys
import time

import click
import globus_sdk

from globus.manager.version import __version__
from globus.manager.api.utils.services import get_dns_client, DNS_SCOPE

from .. import (
    ConfigFile,
    Database,
    EndpointManager,
    GridFTPService,
    GCSManager,
    ServiceManager,
    WebServerManager,
    log as logger,
)
from globus.manager.utils.platform_info import (
    PlatformInfo,
    UnsupportedPlatformException
)

RETRIES = 15
dns_name_file = '/var/lib/globus-connect-server/dns-name.txt'
endpoint_uuid_file = '/var/lib/globus-connect-server/endpoint-uuid.txt'


def register_domain(server_name, auth_client, environment='Production',
                    force=False):
    # Create a new domain and register it with auth
    domain = None
    if os.path.exists(dns_name_file) and not force:
        logger.debug("Checking for existing domain in file {0}".format(
            dns_name_file))
        with open(dns_name_file, 'r') as f:
            domain = f.read().strip()
            if domain == '':
                domain = None
            else:
                logger.debug("Using domain {0}".format(domain))

    if domain is None:
        dns_authorizer = globus_sdk.ClientCredentialsAuthorizer(
            auth_client, DNS_SCOPE)
        dns_client = get_dns_client(
            authorizer=dns_authorizer, environment=environment.lower())
        domain_response = dns_client.create_domain(cname=server_name)
        domain = domain_response.data['domain']
        logger.debug("New domain {0}".format(domain))

        logger.debug("Waiting for domain propagation")
        for attempt in range(1, RETRIES):
            time.sleep(30)
            try:
                # Verify DNS has propagated
                logger.debug("Attempt {0}".format(attempt))
                socket.getaddrinfo(domain, 0)
                break
            except socket.gaierror as e:
                if attempt + 1 < RETRIES:
                    logger.warning(
                        "DNS for {} not resolving, sleeping and trying again"
                        .format(domain))
                else:
                    raise e

        logger.debug("DNS Propagated to us: registering domain with auth")

        for attempt in range(1, RETRIES):
            try:
                auth_response = auth_client.post(
                    '/v2/api/clients/{client_id}/fqdns'.format(
                        client_id=auth_client.client_id),
                    json_body={'fqdn': domain})
                logger.debug("Registered dns with auth returned {0}".format(
                    auth_response.http_status))
                assert auth_response.http_status == 201
                with open(dns_name_file, 'w') as f:
                    f.write(domain + "\n")
                break
            except globus_sdk.exc.AuthAPIError as e:
                logger.debug("Error registering domain: {0} {1}".format(
                    e.http_status, str(e)))
                time.sleep(30)
            except Exception as e:
                logger.debug("Error registering domain: {0}".format(e))
                if attempt + 1 < RETRIES:
                    logger.debug(str(e))
                    time.sleep(30)
                else:
                    raise e

    return domain


def get_certificate(domain, environment=None, force=False):
    cert, key, chain = None, None, None

    for tries in range(1, 10):
        try:
            # Run as gcsweb so it can own/renew certs
            def preexec_fn():
                os.setgid(grp.getgrnam("gcsweb").gr_gid)
                os.setuid(pwd.getpwnam("gcsweb").pw_uid)
            cmd = [
                '/opt/globus/bin/certbot_wrapper',
                '--domain',
                domain,
            ]

            certbot = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, preexec_fn=preexec_fn)
            out, err = certbot.communicate()

            for line in out.decode('utf8').split("\n"):
                if line == '':
                    continue
                name, val = line.split("=", 1)
                if name == 'cert':
                    cert = val
                elif name == 'key':
                    key = val
                elif name == 'chain':
                    chain = val
            if cert is None or key is None or chain is None:
                raise Exception("Letsencrypt error")
            return cert, key, chain
        except KeyError:
            logger.error("Unknown user gcsweb")
            sys.exit(1)
    logger.error("Unable to obtain certificate. DNS or firewall issue")
    sys.exit(1)


@click.command(
    'globus-connect-server-setup',
    short_help='Set up Globus Connect Server services',
    help='''\
The globus-connect-server-setup command performs the initial setup
of the GCS-related services. It creates a GCS endpoint, and then
enables the services needed to interact with the endpoint.

Before running this command, update the /etc/globus-connect-server.conf
file to, at a minimum, contain values for

- [Globus].ClientId

- [Globus].ClientSecret

- [Endpoint].Name

- [LetsEncrypt].Email

- [LetsEncrypt].AgreeToS''')
@click.option(
    '--config-file', '-c',
    help='Use CONFIG_FILE instead of /etc/globus-connect-server.conf',
    metavar='CONFIG_FILE',
    default='/etc/globus-connect-server.conf',
    type=click.Path(readable=True))
@click.version_option(
    version=__version__,
    prog_name='globus-connect-server-setup')
@click.option(
    '--verbose', '-v',
    help='Print more information about tasks',
    is_flag=True)
@click.option(
    '--root', '-r',
    help='Add ROOT as the directory prefix for the configuration',
    metavar='ROOT',
    default='/',
    type=click.Path(dir_okay=True, file_okay=False))
@click.option(
    '--force', '-f',
    help='Ignore errors and existing configuration',
    is_flag=True)
@click.option(
    '--log', '-l',
    help='Log execution information to LOG',
    metavar='LOG',
    default='/var/log/globus-connect-server.log',
    type=click.Path(writable=True, exists=False))
@click.option(
    '--endpoint-uuid', '-u',
    help='Modify the endpoint with the specified UUID',
    metavar='UUID')
@click.pass_context
def main(ctx, config_file, verbose, root, force, log, endpoint_uuid):
    handler = logging.StreamHandler()
    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        handler.setLevel(logging.WARNING)
    logger.addHandler(handler)

    log_path = log
    if root != '/':
        log_path = os.path.join(root, log)

    try:
        old_umask = os.umask(stat.S_IRWXG | stat.S_IRWXO)
        log_file = logging.FileHandler(log_path)
        log_file.setLevel(logging.DEBUG)
        log_file.setFormatter(
            logging.Formatter(
                "%(asctime)s: %(levelname)s: %(message)s",
                datefmt='%Y-%m-%dT%H:%M:%S'))
        logger.addHandler(log_file)
        os.umask(old_umask)
    except Exception:
        pass

    supported_platforms = ['redhat', 'debian']
    if PlatformInfo().platform not in supported_platforms:
        click.echo(str(UnsupportedPlatformException()))
        ctx.exit(1)
    try:
        conf = ConfigFile(config_file=config_file, root=root)
    except Exception as e:
        click.echo('Error parsing configuration: {}'.format(str(e)))
        ctx.exit(1)
    os.environ['GLOBUS_SDK_ENVIRONMENT'] = conf.Globus.Instance().lower()
    os.umask(stat.S_IWGRP | stat.S_IWOTH)
    socket.setdefaulttimeout(300)

    client_id = conf.Globus.ClientId()
    secret = conf.Globus.ClientSecret()

    required_options = {
        '[Globus].ClientId': conf.Globus.ClientId(),
        '[Globus].ClientSecret': conf.Globus.ClientSecret(),
        '[LetsEncrypt].Email': conf.LetsEncrypt.Email(),
        '[LetsEncrypt].AgreeToS': conf.LetsEncrypt.AgreeToS(),
    }
    for o in required_options.keys():
        if not required_options[o]:
            click.echo('Invalid value for configuration option {}'.format(o))
            ctx.exit(1)

    if (
            not force
            and os.path.exists(endpoint_uuid_file)
            and not os.path.exists(dns_name_file)):
        logger.warn(
            '''\
It looks like this endpoint has been configured with GCS 5.0.
If this is true, please contact support@globus.org for help migrating.

If this is not the case, run with the --force option to reconfigure''')
        sys.exit(1)

    if force:
        logger.debug('Stopping GridFTP and Web services')
        if ServiceManager.status(GridFTPService.service):
            ServiceManager.stop(GridFTPService.service)
        if ServiceManager.status(WebServerManager.service):
            ServiceManager.stop(WebServerManager.service)

        logger.debug('Removing old GCS GridFTP configuration')
        for gridftp_conf in glob.glob('/etc/gridftp.d/globus-connect-server*'):
            # Keep file generated by the native package
            if (os.path.basename(gridftp_conf)
                    == 'globus-connect-server-package'):
                continue
            if os.path.islink(gridftp_conf):
                linkedfile = os.readlink(gridftp_conf)
                if os.path.exists(linkedfile):
                    os.remove(linkedfile)
            os.remove(gridftp_conf)

        logger.debug('Removing old GCS Manager Apache configuration')
        WebServerManager.site_disable('gcsmanager')
        WebServerManager.site_remove('gcsmanager')
        WebServerManager.site_disable('gcs_manager')
        WebServerManager.site_remove('gcs_manager')
        if ServiceManager.status('gcs_manager'):
            ServiceManager.stop('gcs_manager')
        if ServiceManager.status('gcs_manager_assistant'):
            ServiceManager.stop('gcs_manager_assistant')

    # load client info into configuration
    database = Database()
    database.configure(
        client_id=client_id,
        client_secret=secret,
        environment=conf.Globus.Instance())

    auth_client = globus_sdk.ConfidentialAppAuthClient(client_id, secret)

    domain = None
    try:
        domain = register_domain(
            conf.Endpoint.ServerName(),
            auth_client,
            environment=conf.Globus.Instance(),
            force=force)
        database.configure(server_name=domain)
    except globus_sdk.exc.AuthAPIError as e:
        if e.http_status == 401:
            logger.error(
                "Error registering DNS domain {0}".format(
                    str(e)))
        elif verbose:
            logger.error(str(e))
        sys.exit(1)

    # enable apache
    for module in WebServerManager.required_modules:
        WebServerManager.module_enable(module)
    for apacheconf in WebServerManager.required_confs:
        WebServerManager.conf_enable(apacheconf)

    if not ServiceManager.status(WebServerManager.service):
        ServiceManager.enable(WebServerManager.service)
        ServiceManager.start(WebServerManager.service)

    ServiceManager.reload(WebServerManager.service)

    # get certificate for GCS manager and gridftp server
    cert, key, chain = get_certificate(
        domain, environment=conf.Globus.Instance(), force=force)
    database.configure(gcs_certificate=cert)

    # enable gridftp server
    gridftp = GridFTPService.create_using_conf(conf, domain=domain, root=root)
    gridftp.setup()

    # create endpoint
    TRANSFER_SCOPE = 'urn:globus:auth:scope:transfer.api.globus.org:all'
    transfer_authorizer = globus_sdk.ClientCredentialsAuthorizer(
        auth_client, TRANSFER_SCOPE)
    api = globus_sdk.TransferClient(
        authorizer=transfer_authorizer)
    # Allow gcsweb to own certificates
    gsi_conf = '/etc/grid-security/gsi.conf'
    with open(gsi_conf, 'r') as old_gsi_conf:
        with open(gsi_conf + '.gcs', 'w') as new_gsi_conf:
            for line in old_gsi_conf:
                if re.match(r'^\s*VHOST_CRED_OWNER\s*=\s*', line) is not None:
                    line = 'VHOST_CRED_OWNER=gcsweb\n'
                new_gsi_conf.write(line)

    os.rename(gsi_conf, gsi_conf + '.gcssave')
    os.rename(gsi_conf + '.gcs', gsi_conf)

    endpoint_manager = EndpointManager.create_using_conf(
        conf=conf,
        api=api, domain=domain,
        endpoint_uuid=endpoint_uuid,
        root=root)
    endpoint_manager.setup()

    # load configuration (endpoint uuid)
    database.configure(host_endpoint_id=endpoint_manager.endpoint_uuid)

    # enable gcs manager
    gcs_manager = GCSManager(domain=domain)
    WebServerManager.site_add(
        gcs_manager.service_name,
        gcs_manager.vhost_configuration(),
        enable=True)
    ServiceManager.restart(WebServerManager.service)
    gcs_manager.setup()

    # enable gcs manager assistant
    ServiceManager.enable('gcs_manager_assistant')
    ServiceManager.start('gcs_manager_assistant')

    ServiceManager.enable('gcs_certbot-renew', periodic=True)
    ServiceManager.start('gcs_certbot-renew', periodic=True)

    print("Deployed GCS Manager: https://{0}".format(domain))
    print("Google OAuth Redirect URL: https://{0}/api/v1/authcallback_google"
          .format(domain))
    print('Created GCS Endpoint "{0}": {1}'.format(
        conf.Endpoint.Name(),
        endpoint_manager.endpoint_uuid))

    print('''\
To enable premium features on this endpoint, a subscription manager must
set the subscription id on endpoint {}'''.format(
        endpoint_manager.endpoint_uuid))
