from __future__ import absolute_import

from .type_handlers import string_handler, enum_handler


class ConfigOption(object):
    """
    This class is used to define configuration options in the configuration
    file.
    """
    def __init__(self, option_name, description,
                 type_handler=None, default_value=None,
                 private=False, comment_out_if_none=True):
        """
        Initialize the configuration option object
        The required parameters are
        - option_name a string value that is the name of the option
        - description a string describing the configuration option

        The optional named parameters are
        - type_handler
          An object that provides validate and parse functions
        - default_value
          A default value if the configuration option is not present or
          has the value None
        - Private
          If True, then the option will not be described in the default
          configuration file
        """
        self.name = option_name
        self.type_handler = type_handler
        if self.type_handler is None:
            self.type_handler = string_handler()
        self.default_value = default_value
        self.description = description
        self.value = default_value
        self.private = private
        self.comment_out_if_none = comment_out_if_none
        if isinstance(self.type_handler, enum_handler):
            for v in self.type_handler.values:
                setattr(self, v.upper().replace(" ", "_"), v)
        self.__doc__ = description

    def set(self, unparsed_value):
        """
        Parse the raw configuration string value into a value of the
        expected type for this configuration option, using the object's
        type_handler to parse the value and the object's default_value if the
        parsed result is None
        """
        if self.type_handler is not None:
            value = self.type_handler.parse(unparsed_value)
        if value is None:
            value = self.default_value
        self.value = value
        return self.value

    def __call__(self):
        return self.value
