"""
Manager for Globus Transfer Endpoints
"""

from __future__ import print_function, absolute_import

import os

from ..utils import log


class EndpointManager(object):
    """
    Class to configure a Globus endpoint
    """
    def __init__(
            self, endpoint_name, require_encryption,
            incoming_port_range, outgoing_port_range,
            api=None, domain=None,
            endpoint_uuid=None, root='/'):
        self.api = api
        self.endpoint_name = endpoint_name
        self.require_encryption = require_encryption
        self.hostname = domain
        self.port = 2811
        self.scheme = 'gsiftp'
        self.incoming_port_range = incoming_port_range
        self.outgoing_port_range = outgoing_port_range
        self.endpoint_id_file = os.path.join(
            root,
            'var/lib/globus-connect-server/endpoint-uuid.txt')
        self.endpoint_uuid = endpoint_uuid
        if self.endpoint_uuid is not None:
            with open(self.endpoint_id_file, "w") as id_file:
                id_file.write("{0}\n".format(self.endpoint_uuid))
        elif os.path.exists(self.endpoint_id_file):
            with open(self.endpoint_id_file) as id_file:
                self.endpoint_uuid = id_file.read().strip()
                if self.endpoint_uuid == "":
                    self.endpoint_uuid = None

    def endpoint_data(self, update=False):
        """
        Return the endpoint data document
        """
        ep_data = {
            "DATA_TYPE": "endpoint",
            "display_name": self.endpoint_name,
            "force_encryption": self.require_encryption,
            'non_functional': True,
        }

        if not update:
            ep_data['public'] = True

        return ep_data

    def server_data(self):
        """
        Return the server document
        """
        result = {
            'DATA_TYPE': 'server',
            'hostname': self.hostname,
            'port': self.port,
            'scheme': self.scheme,
            'incoming_data_port_start': self.incoming_port_range[0],
            'incoming_data_port_end': self.incoming_port_range[1],
        }
        if self.outgoing_port_range is not None:
            result['outgoing_data_port_start'] = self.outgoing_port_range[0]
            result['outgoing_data_port_end'] = self.outgoing_port_range[1]
        return result

    def create_endpoint(self):
        """
        Create a new endpoint, with an initial server
        """
        endpoint_data = self.endpoint_data()
        endpoint_data['DATA'] = [self.server_data()]
        create_result = self.api.create_endpoint(endpoint_data)

        return create_result

    def update_endpoint(self, endpoint_id):
        """
        Update the endpoint document for a particular endpoint_id
        """
        update_result = self.api.update_endpoint(
            endpoint_id, self.endpoint_data(True))
        if update_result['code'] == 'Updated':
            return self.api.get_endpoint(endpoint_id).data

    def setup(self, reset=False, force=False):
        """
        Adds this gridftp server to the endpoint named in the configuration
        file. If force=True is passed, then the endpoint is deleted prior
        to binding this gridftp server. If reset=True is passed, then
        all other GridFTP servers will be removed from this endpoint before
        adding this one.
        """
        endpoint = None
        if self.endpoint_uuid is not None:
            endpoint = self.api.get_endpoint(self.endpoint_uuid)

        if endpoint is not None and 'id' not in endpoint:
            endpoint = None
            self.endpoint_uuid = None

        if force:
            try:
                if endpoint is not None and 'id' in endpoint:
                    log.debug("Removing old endpoint definition")
                    self.api.delete_endpoint(endpoint['id'])
            except Exception:
                pass

            endpoint = None
            self.endpoint_uuid = None

        if self.endpoint_uuid is not None:
            endpoint = self.update_endpoint(self.endpoint_uuid)
            updated = False

            endpoint_data = endpoint.get("DATA")
            if endpoint_data is not None:
                for server in endpoint_data:
                    if reset:
                        self.api.delete_endpoint_server(
                            endpoint['id'],
                            server['id'])
                    elif (server['hostname'] == self.hostname
                            and server['port'] == self.port):
                        self.api.update_endpoint_server(
                            endpoint['id'],
                            server['id'],
                            self.server_data())
                        updated = True
                        break
            if not updated:
                self.api.add_endpoint_server(
                    endpoint['id'],
                    self.server_data())
        else:
            create_result = self.create_endpoint()
            self.endpoint_uuid = str(create_result.get('id'))
            with open(self.endpoint_id_file, "w") as id_file:
                id_file.write("{0}\n".format(create_result['id']))

    def cleanup(self, delete_endpoint=False):
        """
        If delete_endpoint is True, remove the endpoint, otherwise, remove
        all servers which point to this configuration's GridFTP server.
        """
        if delete_endpoint:
            if self.endpoint_uuid is not None:
                log.debug("Removing old endpoint definition")
                try:
                    self.api.delete_endpoint(self.endpoint_uuid)
                except Exception:
                    pass
                if os.path.exists(self.endpoint_id_file):
                    old_endpoint_uuid = None
                    with open(self.endpoint_id_file) as id_file:
                        old_endpoint_uuid = id_file.read().strip()
                        if old_endpoint_uuid == "":
                            old_endpoint_uuid = None
                    if self.endpoint_uuid == old_endpoint_uuid:
                        os.remove(self.endpoint_id_file)
                self.endpoint_uuid = None
                return

        endpoint = None

        if self.endpoint_uuid is not None:
            endpoint = self.api.get_endpoint(self.endpoint_uuid).data

        if endpoint is None:
            return

        for server in endpoint['DATA']:
            hostname = server['hostname']
            port = server['port']
            server_id = server['id']
            if hostname == self.hostname and port == self.port:
                self.api.delete_endpoint_server(self.endpoint_uuid, server_id)

    @staticmethod
    def create_using_conf(
            conf=None, api=None, domain=None, endpoint_uuid=None, root='/'):
        return EndpointManager(
            conf.Endpoint.Name(),
            conf.GridFTP.RequireEncryption(),
            conf.GridFTP.IncomingPortRange(),
            conf.GridFTP.OutgoingPortRange(),
            api=api, domain=domain,
            endpoint_uuid=endpoint_uuid,
            root=root)
