from __future__ import print_function, absolute_import

import os
try:
    import distro
except Exception:
    import platform as distro
import platform

from ..service_manager import ServiceManager
from ..utils import (
    HostInfo,
    log,
    write_and_link_file,
)
from globus.manager.version import __version__ as version


PRIVATE_DATA_WARNING = """
******************************************************************************
WARNING: Your GridFTP server is behind a NAT, but the Server name resolves
to a private IP address. This probably won't work correctly with Globus.
To remedy, set the DataInterface option in the [GridFTP] section of the
globus-connect-server.conf file to the public IP address of this GridFTP
server
******************************************************************************
"""

TRANSFER_ADDR = '54.237.254.192/29'


class GridFTPService(object):
    service = "globus-gridftp-server"
    """
    Class to configure a GridFTP server
    """
    def __init__(
            self,
            incoming_port_range,
            outgoing_port_range,
            allow_udt,
            environment,
            data_interface,
            require_encryption,
            extra_arg,
            domain,
            service_manager=ServiceManager,
            root='/'):
        self.server = domain
        self.service_manager = service_manager
        self.ipc_uds_path = os.path.join(
            root,
            'var/run/globus-connect-server/ipc')
        self.control_uds_path = os.path.join(
            root,
            'var/run/globus-connect-server/control')
        self.trusted_ca_dir = os.path.join(
            root,
            'var/lib/globus-connect-server/grid-security/certificates')
        self.vhost_dir = os.path.join(
            root,
            os.path.expanduser('~gcsweb/etc/letsencrypt/live'))
        self.etc_gridftp_d = os.path.join(root, 'etc/gridftp.d')
        self.var_gridftp_d = os.path.join(
            root, 'var/lib/globus-connect-server/gridftp.d')
        self.sharing_dn = '/'.join([
            '',
            'C=US',
            'O=Globus Consortium',
            'OU=Globus Online',
            'OU=Transfer User',
            'CN=__transfer__',
        ])
        self.incoming_range = incoming_port_range
        self.outgoing_range = outgoing_port_range
        self.allow_udt = allow_udt
        self.gcsdb = os.path.join(
            root,
            os.path.expanduser('~gcsweb/gcs.db'))
        self.allow_from = None
        if environment.lower() == 'production':
            self.allow_from = TRANSFER_ADDR
        self.data_interface = data_interface
        if self.data_interface is None:
            if HostInfo.is_ec2():
                self.data_interface = HostInfo.public_ip()
            else:
                self.data_interface = domain

        self.require_encryption = require_encryption
        self.extra_arg = extra_arg

    def setup(self, force=False):
        if HostInfo.is_private_ip(self.data_interface):
            log.warn(PRIVATE_DATA_WARNING)

        conf_file_name, conf_link_name = self.__config_names()
        config_options = {
            'version_tag GCS-{}': version,
            'usage_stats_id {}': self.usage_stats_id(),
            'auth_level 19': True,
            'port_range {0[0]},{0[1]}': self.incoming_range,
            '$GLOBUS_TCP_SOURCE_RANGE {0[0]},{0[1]}': self.outgoing_range,
            'data_interface {}': self.data_interface,
            'combined_data_node 1': True,
            'ipc_uds_path {}': self.ipc_uds_path,
            'control_uds_path {}': self.control_uds_path,
            'ipc_port {}': -1,
            'control_tls 1': True,
            'allow_udt 1': self.allow_udt or None,
            'allow_from {}': self.allow_from,
            'threads 1': True,
            'encrypt_data 1': self.require_encryption or None,
            'config_db "{}"': self.gcsdb,
            'sharing_dn "{}"': self.sharing_dn,
            '$X509_CERT_DIR "{}"': self.trusted_ca_dir,
            '$X509_VHOST_CRED_DIR "{}"': self.vhost_dir,
            'log_audit {}': '/var/log/gridftp-audit.log',
            'log_single {}': '/var/log/gridftp.log',
            'log_level {}': 'ERROR,WARN',
            '{}': self.extra_arg,
        }
        write_and_link_file(
            conf_file_name, conf_link_name, "\n".join(
                [opt.format(arg) for opt, arg in config_options.items()
                 if arg is not None] + ['']))
        self.service_manager.enable(self.service)
        if self.service_manager.status(self.service):
            self.service_manager.restart(self.service)
        else:
            self.service_manager.start(self.service)

    @classmethod
    def usage_stats_id(cls):
        dist_string = "-".join([
            d
            for d in list(distro.linux_distribution(
                full_distribution_name=False))[0:2] if d != ""]
            + [a for a in list(platform.architecture())[0:1] if a != ''])
        usid = 'GCS-{}'.format(version)
        if dist_string != '':
            usid += "+{}".format(dist_string)
        return usid

    def cleanup(self, **kwargs):
        self.service_manager.stop(self.service)
        self.service_manager.disable(self.service)

        for name in os.listdir(self.etc_gridftp_d):
            if (name.startswith("globus-connect-server")
                    and name != 'globus-connect-server-package'):
                os.remove(os.path.join(self.etc_gridftp_d, name))
        for name in os.listdir(self.var_gridftp_d):
            os.remove(os.path.join(self.var_gridftp_d, name))

    def __config_names(self, name=None, clean_link=True):
        if name is not None:
            longname = "globus-connect-server-" + name
        else:
            longname = "globus-connect-server"

        conf_file_name = os.path.join(self.var_gridftp_d, longname)
        conf_link_name = os.path.join(self.etc_gridftp_d, longname)

        if clean_link:
            try:
                os.remove(conf_link_name)
            except OSError:
                pass

        return (conf_file_name, conf_link_name)

    @staticmethod
    def create_using_conf(
            conf, domain, service_manager=ServiceManager, root='/'):
        return GridFTPService(
            conf.GridFTP.IncomingPortRange(),
            conf.GridFTP.OutgoingPortRange(),
            conf.GridFTP.AllowUDT(),
            conf.Globus.Instance(),
            conf.GridFTP.DataInterface(),
            conf.GridFTP.RequireEncryption(),
            conf.GridFTP.ExtraArg(),
            domain,
            service_manager=service_manager,
            root=root)
