"""v5.1 migration

Schema/data migration for the storage_type to storage_systems refactor.

Revision ID: eecf5cdfabce
Revises: dba9a86d5a31
Create Date: 2017-10-09 14:32:09.612783

"""
from alembic import op
import sqlalchemy as sa

import json
import uuid

# revision identifiers, used by Alembic.
revision = 'eecf5cdfabce'
down_revision = 'dba9a86d5a31'
branch_labels = None
depends_on = None

GDRIVE_STORAGE_TYPE_ID = '976cf0cf-78c3-4aab-82d2-7c16adbcc281'


def upgrade():
    conn = op.get_bind()

    op.rename_table(u'storage_type', u'temp_storage_type')
    op.create_table(
        'storage_type',
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('id', sa.TEXT(), nullable=False),
        sa.Column('display_name', sa.TEXT(), nullable=False),
        sa.Column('enabled', sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('display_name'))

    conn.execute(
        """
        insert into storage_type(created_at, updated_at, id, display_name,
        enabled) select created_at, updated_at, id, display_name, enabled
        from temp_storage_type""")

    conn.execute(
        """update storage_type set enabled = 1 where display_name = 'POSIX'""")

    op.create_table(
        'storage_system',
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('id', sa.TEXT(), nullable=False),
        sa.Column('display_name', sa.TEXT(), nullable=False),
        sa.Column('credential', sa.TEXT(), nullable=True),
        sa.Column('storage_policies', sa.TEXT(), nullable=True),
        sa.Column('deleted', sa.Boolean(), server_default='0', nullable=False),
        sa.Column('storage_type_id', sa.TEXT(), nullable=False),
        sa.ForeignKeyConstraint(['storage_type_id'], ['storage_type.id']),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('display_name'))

    op.create_index(
        op.f('ix_storage_system_deleted'),
        'storage_system', ['deleted'], unique=False)

    select_gdrive_text = """
        select created_at, updated_at, id, credential,
        storage_policies from temp_storage_type where id = :id
        and credential is not null
        """
    select_gdrive = sa.text(select_gdrive_text)

    insert_text = """
        insert into storage_system(created_at, updated_at, id,
        display_name, credential, storage_policies, deleted,
        storage_type_id)
        VALUES(:created, :updated, :id, :display_name, :credential,
        :storage_policies, :deleted, :storage_type_id)
        """
    insert = sa.text(insert_text)

    has_gdrive = conn.execute(select_gdrive, id=GDRIVE_STORAGE_TYPE_ID).first()

    if has_gdrive:
        (_created, _updated, _id, _credential, _storage_policies) = has_gdrive
        gdrive_storage_system_id = str(uuid.uuid4())

        _storage_policies = json.loads(_storage_policies)
        new_storage_policies = {
            'private': {
                'process_user': 'gcsweb',
                'load_dsi_module': 'google_drive'
            },
            'public': {
                 'root': '/',
            }
        }

        for k, v in _storage_policies.items():
            new_storage_policies['public'][k] = v

        new_storage_policies = json.dumps(new_storage_policies)

        conn.execute(
            insert,
            created=_created, updated=_updated, id=gdrive_storage_system_id,
            display_name='gdrive_migration_50', credential=_credential,
            storage_policies=new_storage_policies, deleted='0',
            storage_type_id=_id)

    op.rename_table(u'user_credential', u'temp_user_credential')
    op.create_table(
        'user_credential',
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('id', sa.TEXT(), nullable=False),
        sa.Column('display_name', sa.TEXT(), nullable=True),
        sa.Column('identity_id', sa.TEXT(), nullable=False),
        sa.Column('storage_credential', sa.TEXT(), nullable=True),
        sa.Column('storage_system_id', sa.TEXT(), nullable=False),
        sa.Column('storage_system_user_id', sa.TEXT(), nullable=False),
        sa.Column('deleted', sa.Integer(), server_default='0', nullable=False),
        sa.Column('invalid', sa.Integer(), server_default='0', nullable=False),
        sa.UniqueConstraint(
            'storage_system_id', 'storage_system_user_id', 'identity_id'),
        sa.ForeignKeyConstraint(
            ['storage_system_id'], ['storage_system.id'], ),
        sa.PrimaryKeyConstraint('id'))

    if has_gdrive:
        conn.execute(
            """
            insert into user_credential(created_at, updated_at, id,
            display_name, identity_id, storage_credential, storage_system_id,
            storage_system_user_id) select created_at, updated_at, id,
            display_name, identity_id, storage_credential, '{}',
            storage_type_user_id from temp_user_credential""".format(
                gdrive_storage_system_id))

    op.create_index(
        op.f('ix_user_credential_storage_system_user_id'),
        'user_credential', ['storage_system_user_id'], unique=False)

    op.rename_table(u'collection', u'temp_collection')
    op.create_table(
        'collection',
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=False),
        sa.Column('id', sa.TEXT(), nullable=False),
        sa.Column('path', sa.TEXT(), nullable=True),
        sa.Column('user_credential_id', sa.TEXT(), nullable=False),
        sa.ForeignKeyConstraint(
            ['user_credential_id'], ['user_credential.id']),
        sa.PrimaryKeyConstraint('id'))
    conn.execute('insert into collection select * from temp_collection')

    op.drop_table(u'temp_collection')
    op.drop_table(u'temp_user_credential')
    op.drop_table(u'temp_storage_type')

    op.create_index(
        op.f('ix_user_credential_deleted'), 'user_credential', ['deleted'],
        unique=False)
    op.create_index(
        op.f('ix_user_credential_invalid'), 'user_credential', ['invalid'],
        unique=False)


def downgrade():
    # downgrading from v5.1 to v5.0 would be a destructive operation
    # in the event that the system had multiple storage systems of a
    # given storage type.
    raise Exception('Downgrading from v5.1 is not supported')
